
#include "losing.h"

typedef struct {uint64 hash; uint64 offset;} tKNOWN;
static tKNOWN *KNOWN_TABLE;
static uint64 KNOWN_MASK;
static uint64 KNOWN_OFFSET[0x1000];
static uint64 known_num;

static uint8 FRONT[4]={0x83,0x37,0xfb,0x26},BACK[4]={0x29,0xae,0x47,0x08};

uint64 known_find(uint64 Z)
{uint64 k=Z&KNOWN_MASK; uint64 z;
 while ((z=KNOWN_TABLE[k].hash))
 {if (z==Z) return KNOWN_TABLE[k].offset; k++; if (k>KNOWN_MASK) k=0;}
 return -1;}

static void known_add_os(uint64 Z,uint64 os)
{uint64 k=Z&KNOWN_MASK; uint64 z;
 while ((z=KNOWN_TABLE[k].hash))
 {if (z==Z) return; k++; if (k>KNOWN_MASK) k=0;}
 KNOWN_TABLE[k].hash=Z; KNOWN_TABLE[k].offset=os;}

static void rebuild_known()
{int u; uint64 sz; tKNOWN *OLD_KNOWN;
 sz=KNOWN_MASK+1; KNOWN_MASK=2*KNOWN_MASK+1; // no need to clear known_num
 OLD_KNOWN=KNOWN_TABLE; KNOWN_TABLE=malloc(sizeof(tKNOWN)*(sz*2));
 for (u=0;u<2*sz;u++) KNOWN_TABLE[u].hash=KNOWN_TABLE[u].offset=0;
 for (u=0;u<sz;u++)
  if (OLD_KNOWN[u].hash) known_add_os(OLD_KNOWN[u].hash,OLD_KNOWN[u].offset);
 free(OLD_KNOWN);}

static void known_add(uint64 Z,uint64 sz,boolean STARTUP)
{uint64 k=Z&KNOWN_MASK; uint64 z;
 while ((z=KNOWN_TABLE[k].hash)) // inc OFFSET when HASH is found on STARTUP
 {if (z==Z) {if (STARTUP) KNOWN_OFFSET[Z&0xfff]+=sz; return;}
  k++; if (k>KNOWN_MASK) k=0;}
 KNOWN_TABLE[k].hash=Z; KNOWN_TABLE[k].offset=KNOWN_OFFSET[Z&0xfff];
 KNOWN_OFFSET[Z&0xfff]+=sz;
 known_num++; if (known_num*2>KNOWN_MASK) rebuild_known();}

static void load_known_index(int u)
{char O[64]; FILE *F; uint64 hash,sz;
 sprintf(O,"KNOWN/iKNOWN0x%03x",u); F=fopen(O,"r"); if (!F) return;
 while (fread(&hash,8,1,F)) {fread(&sz,8,1,F); known_add(hash,sz,TRUE);}
 fclose(F);}

static void load_known_indices()
{int i; for (i=0;i<0x1000;i++) load_known_index(i);}

#define DONT_USE_KNOWN_LOCK TRUE
#include <unistd.h>
void init_known(boolean WRITEABLE) // KNOWN.lock is lock file for KNOWN data
{int i,u; uint64 sz; known_num=0; for (i=0;i<0x1000;i++) KNOWN_OFFSET[i]=0;
 sz=0x100000; KNOWN_MASK=sz-1; KNOWN_TABLE=malloc(sizeof(tKNOWN)*sz);
 for (u=0;u<sz;u++) KNOWN_TABLE[u].hash=KNOWN_TABLE[u].offset=0;
 load_known_indices(); if (DONT_USE_KNOWN_LOCK || !WRITEABLE) return;
 FILE *F; F=fopen("KNOWN.lock","r");
 if (F) {printf("KNOWN.lock!\n"); fflush(stdout); exit(1);}
 F=fopen("KNOWN.lock","w"); fprintf(F,"PID is %d\n",getpid()); fclose(F);}

void new_known(uint64 Z,uint8 *data,uint64 nb)
{char O[64]; FILE *F; uint64 bc=nb+16; if (known_find(Z)!=-1) return;
 sprintf(O,"KNOWN/KNOWN0x%03llx",Z&0xfff); F=fopen(O,"a"); if (!F) return;
 fwrite(FRONT,1,4,F); fwrite(&Z,8,1,F);
 fwrite(data,1,nb,F); fwrite(BACK,1,4,F); fclose(F); known_add(Z,bc,FALSE);
 sprintf(O,"KNOWN/iKNOWN0x%03llx",Z&0xfff); F=fopen(O,"a");
 if (!F) return; fwrite(&Z,8,1,F); fwrite(&bc,8,1,F); fclose(F);}

WIN_STRUCT* WintreeKnown(uint64 Z)
{uint64 hash,u,d,sz,os=known_find(Z); uint16 s; uint8 tmp[4];
 char O[64]; FILE *F; WIN_STRUCT *WS; WIN_NODE *W; if (os==-1) return NULL;
 sprintf(O,"KNOWN/KNOWN0x%03llx",Z&0xfff); F=fopen(O,"r");
 if (!F) {printf("File failure %s\n",O); return NULL;}
 fseek(F,os,SEEK_SET); fread(tmp,1,4,F);
 for (u=0;u<4;u++) if (FRONT[u]!=tmp[u])
 {printf("BAD header %016llx %lld\n",Z,os); fclose(F); return NULL;}
 fread(&hash,8,1,F);
 if (hash!=Z) {printf("BAD hash %016llx %lld\n",Z,os); fclose(F); return NULL;}
 WS=malloc(sizeof(WIN_STRUCT)); WS->W=malloc(4096); W=WS->W;
 fread(&(W[0].data),4,1,F); fread(&(W[0].move),2,1,F);
 d=W[0].data&0x3fffffff; s=W[0].move; WS->MOVE_LIST=malloc(s*sizeof(uint16));
 // printf("Record size %d\n",d); printf("Move list size %d\n",s);
 for (u=0;u<s;u++) fread(&(WS->MOVE_LIST[u]),2,1,F);
 sz=((uint64) d)*sizeof(WIN_NODE); WS->W=realloc(WS->W,sz+65536);
 for (u=1;u<d;u++)
 {fread(&(WS->W[u].data),4,1,F); fread(&(WS->W[u].move),2,1,F);}
 fread(tmp,1,4,F);
 for (u=0;u<4;u++) if (BACK[u]!=tmp[u])
 {printf("BAD footer %016llx %lld\n",Z,os); free(WS->MOVE_LIST);
  free(W); free(WS); fclose(F); return NULL;} fclose(F); return WS;}
