
#include "losing.h"
#include <math.h>

PN_NODE *TREE;  typePOS POS[1];

FILE* WRITE[64]; FILE* READ[64]; int fd_in[64];
int NUM_AVAIL[64],MAX_AVAIL[64]; boolean IS_ALIVE[64],IS_DEAD[64];
char REGISTER_STRING[64][1024];
int QUEUED=0,WORKING=0; double WIDE_POWER=2.0,WIDE_LIMIT=30.0;

#define IS_DONE 0
#define IN_QUEUE 1
#define IS_WORKING 2
#define IGNORE_IT 3
#define LOGGING TRUE
#define BUG_TRACKING FALSE

FILE *BUGFILE=NULL; FILE *FLOGFILE=NULL;

static int NUM_SLAVES=0;
static boolean QUIT=FALSE,PAUSE=FALSE;
#define QUEUE_SIZE 0x10000
static uint8 TABLE_STATUS[QUEUE_SIZE];
static uint32 TABLE_NODE[QUEUE_SIZE];
static uint32 TIME;

#define TB5RAT 1048576.0
#define TEMPO 1.0
#define TEMPO2 (TEMPO*TEMPO)

////////////////////////////////////////////////////////////////////////

#include <unistd.h>
static boolean TryInput()
{fd_set fd[1]; struct timeval tv[1]; FD_ZERO(fd); FD_SET(STDIN_FILENO,fd);
 tv->tv_sec=tv->tv_usec=0; return select(STDIN_FILENO+1,fd,NULL,NULL,tv)>0;}

#include <time.h>
static void NANO(int x)
{struct timespec TS[1]; TS->tv_sec=0; TS->tv_nsec=x; nanosleep(TS,NULL);}

////////////////////////////////////////////////////////////////////////

#define FLOAT_INF 9007199254740992.0
static boolean node_ratio(uint32 NODE)
{float r,x=TREE[NODE].rat; uint32 n;
 if (!(n=TREE[NODE].child)) return FALSE; r=TREE[n].rat;
 if (x==0.0 || x==FLOAT_INF) return FALSE;
 while ((n=TREE[n].sibling)) r=MIN(r,TREE[n].rat);
 if (r==FLOAT_INF) TREE[NODE].rat=0.0;
 else if (r==0.0) TREE[NODE].rat=FLOAT_INF; else TREE[NODE].rat=TEMPO2/r;
 return (x!=TREE[NODE].rat);}

static void node_ratio_backtrack(uint32 NODE)
{uint32 k,n; if (!NODE) return; n=TREE[NODE].loop?TREE[NODE].trans:NODE;
 if (node_ratio(n)) node_ratio_backtrack(TREE[n].parent);
 k=TREE[n].loop; if (!k) return;
 while (k!=n && k) // checking k not really necessary here (but is below)
 {float r=TREE[k].rat; TREE[k].rat=TREE[n].rat;
  if (r!=TREE[n].rat) node_ratio_backtrack(TREE[k].parent); k=TREE[k].loop;}}

////////////////////////////////////////////////////////////////////////

typedef struct {uint32 n,good,bad;} SORTER;
static int sort_compare(const void *x, const void *y) // increasing good
{SORTER A=*((SORTER*) x); SORTER B=*((SORTER*) y);
 if (A.good<B.good) return -1; if (A.good>B.good) return 1;
 if (A.bad>B.bad) return -1; if (A.bad<B.bad) return 1;
 return 0;}

void sort_childrenA(uint32 node)
{SORTER S[256]; int i,u=0; uint32 n=TREE[node].child;
 S[u].bad=TREE[n].bad; S[u].good=TREE[n].good; S[u++].n=n;
 while (TREE[n].sibling)
 {n=TREE[n].sibling; S[u].bad=TREE[n].bad; S[u].good=TREE[n].good; S[u++].n=n;}
 qsort(S,u,sizeof(SORTER),sort_compare);
 n=S[0].n; TREE[node].child=S[0].n;
 for (i=1;i<u;i++) {TREE[n].sibling=S[i].n; n=S[i].n;} TREE[n].sibling=0;}

#define LOSS_TWIST 0x913bb468de01037fULL
#define WIN_TWIST 0xc066034f796185ebULL
static uint32 IsKnownResult(uint64);
static void InstallKnownResult(uint64,uint32);
static void backtrack_loopA(uint32);

static HASH *ALL_HASH=NULL;
static uint32 ALL_HASH_MASK;

static void win_installer(uint32 node)
{uint32 u,k; uint64 z,Z; InstallKnownResult(TREE[node].hash^WIN_TWIST,node);
 TREE[node].rat=FLOAT_INF; node_ratio_backtrack(TREE[node].parent);
 TREE[node].loop=0; TREE[node].trans=node;
 TREE[node].bad=0; TREE[node].good=MY_INFoo;
 if (!TREE[node].size) TREE[node].size=TIME--;
 if (BUG_TRACKING) fprintf(BUGFILE,"WIN %d %llx\n",node,TREE[node].hash);
 Z=TREE[node].hash; k=Z&ALL_HASH_MASK;
 while ((z=ALL_HASH[k].key))
 {u=ALL_HASH[k].node; k++; if (k==(ALL_HASH_MASK+1)) k=0;
  if ((z==Z) && u!=node)
  {if (!TREE[u].good) continue; TREE[u].bad=0; TREE[u].good=MY_INFoo;
   TREE[u].flags|=4; TREE[u].trans=node; TREE[u].loop=0; TREE[u].child=0;
   if (!TREE[u].size) TREE[u].size=TIME--;
   if (BUG_TRACKING) fprintf(BUGFILE,"ALSOw %d %d %llx\n",u,node,TREE[u].hash);
   backtrack_loopA(TREE[u].parent);
   TREE[u].rat=FLOAT_INF; node_ratio_backtrack(TREE[u].parent);}}}

// should change all trans *before* tracing back?

static void loss_installer(uint32 node)
{uint32 u,k; uint64 z,Z; InstallKnownResult(TREE[node].hash^LOSS_TWIST,node);
 TREE[node].rat=0.0; node_ratio_backtrack(TREE[node].parent);
 TREE[node].loop=0; TREE[node].trans=node;
 TREE[node].good=0; TREE[node].bad=MY_INFoo;
 if (!TREE[node].size) TREE[node].size=TIME--;
 if (BUG_TRACKING) fprintf(BUGFILE,"LOSS %d %llx\n",node,TREE[node].hash);
 Z=TREE[node].hash; k=Z&ALL_HASH_MASK;
 while ((z=ALL_HASH[k].key))
 {u=ALL_HASH[k].node; k++; if (k==(ALL_HASH_MASK+1)) k=0;
  if ((z==Z) && u!=node)
  {if (!TREE[u].bad) continue; TREE[u].good=0; TREE[u].bad=MY_INFoo;
   TREE[u].flags|=4; TREE[u].trans=node; TREE[u].loop=0; TREE[u].child=0;
   if (!TREE[u].size) TREE[u].size=TIME--;
   if (BUG_TRACKING) fprintf(BUGFILE,"ALSOl %d %d %llx\n",u,node,TREE[u].hash);
   backtrack_loopA(TREE[u].parent);
   TREE[u].rat=0.0; node_ratio_backtrack(TREE[u].parent);}}}

// bad thing: pos turns to WIN based on hash/trans...
// but it was already a LOSS due to repetition!

static boolean update_node(uint32 node)
{uint32 n,v,w=TREE[node].good,l=TREE[node].bad;
 if (BUG_TRACKING) fprintf(BUGFILE,"UPDATE %d %d %d\n",node,w,l);
 if (!TREE[node].child) return TRUE; sort_childrenA(node);
 if (!w || !l) return FALSE; // think this also prevents loops
 n=TREE[node].child; v=TREE[n].bad; TREE[node].bad=TREE[n].good;
 while (TREE[n].sibling)
 {n=TREE[n].sibling; v+=TREE[n].bad; if (v>MY_INFoo) v=MY_INFoo;}
 if (w && !v && !(TREE[node].flags&16) &&
     !IsKnownResult(TREE[node].hash^LOSS_TWIST)) loss_installer(node);
 if (l && !TREE[node].bad && (TREE[node].flags&16) &&
     !IsKnownResult(TREE[node].hash^WIN_TWIST)) win_installer(node);
 TREE[node].good=v; if (v==w && l==TREE[node].bad) return FALSE; return TRUE;}
// Do NOT install White losses or Black wins, could be reps!

static void backtrack_loopA(uint32 node)
{uint32 n,k; if (!node) return; n=TREE[node].loop?TREE[node].trans:node;
 if (update_node(n)) backtrack_loopA(TREE[n].parent);
 k=TREE[n].loop; if (!k) return;
 while (k!=n && k) // note that loop could be set to zero!
 {uint32 g=TREE[k].good,b=TREE[k].bad;
  TREE[k].good=TREE[n].good; TREE[k].bad=TREE[n].bad;
  if (g!=TREE[n].good || b!=TREE[n].bad) backtrack_loopA(TREE[k].parent);
  k=TREE[k].loop;}}

// update_node() returns TRUE with no chld, while node_ratio() returns FALSE

////////////////////////////////////////////////////////////////////////

HASH *BOOK_HASH;
uint32 BOOK_HASH_MASK;

static void remake_all_hash()
{uint32 k,u; HASH *NEW_HASH; uint32 osz=ALL_HASH_MASK+1,sz=osz*2;
 if (BUG_TRACKING){fprintf(BUGFILE,"REMAKE_ALL_HASH %d\n",sz);fflush(BUGFILE);}
 ALL_HASH_MASK=sz-1; NEW_HASH=malloc(sizeof(HASH)*sz);
 for (u=0;u<sz;u++) NEW_HASH[u].key=0;
 for (u=0;u<osz;u++)
 {if (!ALL_HASH[u].key) continue; k=ALL_HASH[u].key&ALL_HASH_MASK;
  while (NEW_HASH[k].key) {k++; if (k==sz) k=0;}
  NEW_HASH[k].key=ALL_HASH[u].key; NEW_HASH[k].node=ALL_HASH[u].node;}
 free(ALL_HASH); ALL_HASH=NEW_HASH;}

static void MakeAllHash(uint32 n,uint64 Z)
{uint32 k=Z&ALL_HASH_MASK;
 while (ALL_HASH[k].key) {k++; if (k==(ALL_HASH_MASK+1)) k=0;}
 ALL_HASH[k].key=Z; ALL_HASH[k].node=n;
 if ((TREE[0].size*3)/2>ALL_HASH_MASK) remake_all_hash();}

static uint32 NewBookHash(uint64 Z,uint32 NODE)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return BOOK_HASH[k].node; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
 BOOK_HASH[k].key=Z; BOOK_HASH[k].node=NODE; return 0;}

static void InstallKnownResult(uint64 Z,uint32 n)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
 BOOK_HASH[k].key=Z; BOOK_HASH[k].node=n;}

static uint32 IsKnownResult(uint64 Z)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return BOOK_HASH[k].node; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
 return FALSE;}

// flags: 1 is TB5, 4 is trans res, 16 is wtm, 64 is exterior res

// #define KRNPPP

#define TB5RAT 1048576.0

static void initial_book_scan()
{uint32 u; for (u=1;u<TREE[0].size;u++)
 {MakeAllHash(u,TREE[u].hash);
  if (TREE[u].flags&4 || TREE[u].flags&64) continue;
  if (TREE[u].flags&1 && TREE[u].rat!=-1 && TREE[u].bad && TREE[u].good)
    TREE[u].rat*=(TREE[u].flags&16)?TB5RAT:1/TB5RAT;
  if (!TREE[u].good && !(TREE[u].flags&16) &&
      !IsKnownResult(TREE[u].hash^LOSS_TWIST))
    InstallKnownResult(TREE[u].hash^LOSS_TWIST,u);
  if (!TREE[u].bad && (TREE[u].flags&16) &&
      !IsKnownResult(TREE[u].hash^WIN_TWIST))
    InstallKnownResult(TREE[u].hash^WIN_TWIST,u);}}

static void InitBookHash(uint32 NODE)
{uint32 n; if (TREE[NODE].flags&4) return;
 if (TREE[NODE].hash!=POS->DYN->HASH)
  printf("HASH prob %d %llx %llx\n",NODE,TREE[NODE].hash,POS->DYN->HASH);
#if 0
 if (TREE[NODE].flags&1 && TREE[NODE].rat!=-1 &&
     TREE[NODE].bad && TREE[NODE].good)
   TREE[NODE].rat*=POS->wtm?TB5RAT:1/TB5RAT;
#endif
 if (TREE[NODE].loop && TREE[NODE].trans!=NODE) return;
 if (POS->DYN->rev==0) NewBookHash(POS->DYN->HASH,NODE); // could be there!
#if 0
 if (!TREE[NODE].good &&!POS->wtm && !IsKnownResult(POS->DYN->HASH^LOSS_TWIST))
   InstallKnownResult(POS->DYN->HASH^LOSS_TWIST,NODE);
 if (!TREE[NODE].bad && POS->wtm && !IsKnownResult(POS->DYN->HASH^WIN_TWIST))
   InstallKnownResult(POS->DYN->HASH^WIN_TWIST,NODE);
#endif
 n=TREE[NODE].child; if (!n) return;
 if (!MakeMove(POS,TREE[n].move)) return; // hmm
 InitBookHash(n); UnmakeMove(POS,TREE[n].move,TRUE);
 while ((n=TREE[n].sibling))
 {if (!MakeMove(POS,TREE[n].move)) continue;
  InitBookHash(n); UnmakeMove(POS,TREE[n].move,TRUE);}}
// should load all won hashes, not just those reachable from root!

static void init_all_ratios()
{uint32 q,i,t,k,c,u,v=0,z=0; float r; uint32 *ARR;
 ARR=malloc(TREE[0].size*sizeof(uint32));
 for (i=1;i<TREE[0].size;i++)
 {k=TREE[i].loop; t=k?TREE[i].trans:i; if (!k || TREE[t].child) continue;
  k=TREE[t].loop; while (k!=t) {TREE[k].rat=TREE[t].rat; k=TREE[k].loop;}}
 for (i=1;i<TREE[0].size;i++)
  if ((!TREE[i].loop || TREE[i].trans==i) && TREE[i].rat==-1) ARR[v++]=i;
 while (TREE[1].rat==-1 && z<1000)
 {u=v; if (0) printf("PASS %d: left %d\n",++z,u); v=0;
  for (i=0;i<u;i++)
  {q=ARR[i]; t=TREE[q].loop?TREE[q].trans:q;
   if (TREE[t].rat!=-1.0) {ARR[v++]=ARR[i]; continue;}
   c=TREE[t].child; if (!c || TREE[c].rat==-1.0) {ARR[v++]=ARR[i]; continue;}
   k=TREE[t].loop; r=FLOAT_INF; r=MIN(r,TREE[c].rat);
   while ((c=TREE[c].sibling))
   {if (TREE[c].rat==-1.0) break; r=MIN(r,TREE[c].rat);}
   if (c) {ARR[v++]=ARR[i]; continue;} if (r==FLOAT_INF) TREE[t].rat=0.0;
   else if (r==0.0) TREE[t].rat=FLOAT_INF; else TREE[t].rat=TEMPO2/r;
   if (k) while (k!=t) {TREE[k].rat=TREE[t].rat; k=TREE[k].loop;}}} free(ARR);}

////////////////////////////////////////////////////////////////////////

static void path_nodeA(uint32 n)
{uint16 M[256]; int k=0;
 while (TREE[n].parent) {M[k++]=TREE[n].move; n=TREE[n].parent;}
 while (k>0) MakeMoveCheck(POS,M[--k],FALSE);} /* necessary to allow reps */

#if 1
void print_path(uint32 n) // utility
{uint16 M[256]; char A[8]; int k=0; printf("%d ",n);
 while (TREE[n].parent) {M[k++]=TREE[n].move; n=TREE[n].parent;}
 while (k>0) printf(" %s",Notate(A,M[--k])); printf("\n");}
#endif

boolean TB_PiSq_score(type_PiSq *PiSq,uint8 *v);

#define EGTB_LIM 6
boolean get_tb_score(uint8 *va)
{int i,n=0; type_PiSq PiSq[1]; int pi,sq; uint64 O;
 if (POS->DYN->ep) return FALSE; for (i=0;i<6;i++) PiSq->sq[i]=PiSq->pi[i]=0;
 PiSq->wtm=POSITION->wtm; O=wBitboardOcc|bBitboardOcc; PiSq->Occupied=O;
 if (POPCNT(O)>EGTB_LIM) return FALSE;
 PiSq->n=0; PiSq->pawn=FALSE; PiSq->Blocked=0;
 while (O)
 {sq=BSF(O); O&=(O-1); pi=POSITION->sq[sq]; PiSq->pi[n]=pi; PiSq->sq[n++]=sq;
  if (pi==wEnumP || pi==bEnumP) PiSq->pawn=TRUE;}
 PiSq->n=n; return TB_PiSq_score(PiSq,va);}

static void egtb5_result(uint32 n)
{uint8 va; boolean win;
 if (!get_tb_score(&va)) {if (0) printf("FAILED\n"); return;}
 if ((va<=2 && !POS->wtm) || (va==1 && POS->wtm)) win=TRUE; else win=FALSE; 
 TREE[n].flags|=1; if ((win && POS->wtm) || (!win && !POS->wtm)) return;
 if (win) {TREE[n].good=MY_INFoo; TREE[n].bad=0;}
 if (!win) {TREE[n].good=0; TREE[n].bad=MY_INFoo;}}

#define TB5RAT 1048576.0
#define REV_LIM 75 /* should be less than REV_LIMIT (95) in gen_moves.c */

static void new_node(int node,int good,int bad,int parent,int mv)
{uint32 c,t; TREE[node].flags=POS->wtm?16:0; TREE[node].hash=POS->DYN->HASH;
 TREE[node].good=good; TREE[node].bad=bad; MakeAllHash(node,POS->DYN->HASH);
 if (POPCNT(POS->OCCUPIED)<=EGTB_LIM) egtb5_result(node);
 TREE[node].parent=parent; TREE[node].sibling=TREE[node].child=0;
 TREE[node].move=mv; TREE[node].size=TREE[node].killer=0;
 TREE[node].trans=node; TREE[node].loop=0; TREE[node].size=0;
 if (!TREE[parent].child) TREE[parent].child=node;
 else {c=TREE[parent].child; while (TREE[c].sibling) c=TREE[c].sibling;
       TREE[c].sibling=node;}
 if (!TREE[node].bad) TREE[node].rat=FLOAT_INF;
 else TREE[node].rat=(TEMPO*TREE[node].good)/TREE[node].bad;
 if (TREE[node].bad && TREE[node].good && TREE[node].flags&1)
 {if (TREE[node].flags&16) TREE[node].rat*=TB5RAT;
  else TREE[node].rat/=TB5RAT;}
 if (POS->wtm && (t=IsKnownResult(TREE[node].hash^WIN_TWIST)))
 {TREE[node].good=MY_INFoo; TREE[node].bad=0;
  TREE[node].flags|=4; TREE[node].size=TIME--;
  TREE[node].rat=FLOAT_INF; TREE[node].trans=t; TREE[node].loop=0;
  if (BUG_TRACKING) fprintf(BUGFILE,"NEWw %d %d %llx\n",node,t,POS->DYN->HASH);
  return;}
 if (!POS->wtm && (t=IsKnownResult(TREE[node].hash^LOSS_TWIST)))
 {TREE[node].bad=MY_INFoo; TREE[node].good=0;
  TREE[node].flags|=4; TREE[node].size=TIME--;
  TREE[node].rat=0.0; TREE[node].trans=t; TREE[node].loop=0;
  if (BUG_TRACKING) fprintf(BUGFILE,"NEWl %d %d %llx\n",node,t,POS->DYN->HASH);
  return;}
 if ((POS->wtm && !bad) || (!POS->wtm && !good)) TREE[node].size=TIME--;
 else if (POS->DYN->rev > REV_LIM)
 {if (POS->wtm) {TREE[node].bad=MY_INFoo; TREE[node].good=0;}
  else {TREE[node].good=MY_INFoo; TREE[node].bad=0;}}
 if (POS->DYN->rev==0 && !IsKnownResult(TREE[node].hash^WIN_TWIST) &&
     !IsKnownResult(TREE[node].hash^LOSS_TWIST) &&
     ((t=NewBookHash(POS->DYN->HASH,node)))) // old info
 {TREE[node].trans=t; TREE[node].rat=TREE[t].rat;
  TREE[node].good=TREE[t].good; TREE[node].bad=TREE[t].bad;
  TREE[node].loop=TREE[t].loop?TREE[t].loop:t; TREE[t].loop=node;}}

static uint32 TreeMove(uint16 mv,uint32 NODE)
{uint32 n=TREE[NODE].child; if (!n) return 0; if (TREE[n].move==mv) return n;
 while ((n=TREE[n].sibling)) {if (TREE[n].move==mv) return n;}
 return 0;}

static uint16 TABLE_PATH[QUEUE_SIZE][256];
static void walk_path(int k) // could make this MakeMoveCheck if allow rep
{int u=0; uint16 mv; while ((mv=TABLE_PATH[k][u++])) MakeMove(POS,mv);}
// but Slave would have to allow that too then

static uint32 align_booktreeA(uint32 NODE,boolean *b)
{typeDYNAMIC *D; *b=FALSE; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && NODE)
 {D++; NODE=TreeMove(D->mv,NODE);
  if (!TREE[NODE].bad || !TREE[NODE].good) *b=TRUE;
  if (TREE[NODE].loop) NODE=TREE[NODE].trans;}
 return NODE;}

static uint32 align_booktreeB(uint32 NODE,boolean *b)
{typeDYNAMIC *D; *b=FALSE; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && NODE)
 {D++; if (TREE[NODE].loop) NODE=TREE[NODE].trans; NODE=TreeMove(D->mv,NODE);
  if (!TREE[NODE].bad || !TREE[NODE].good) *b=TRUE;} return NODE;}

uint8 *BYTE_ARR;

static int walk_to_pos(uint32 node,char *ml,int *move)
{uint32 c; char S[8]; int i,mv; // path_node can induce reps
 init_position(POS,"startpos"); path_nodeA(node);
 c=strlen(ml); for (i=0;i<c;i++) if (ml[i]=='\n') ml[i]=0;
 if (!ml[0]) return node;
 while (TRUE) 
 {sscanf(ml,"%s",S); if (ml[strlen(S)]==0) break; ml=ml+strlen(S)+1;
  mv=get_move(POS,S); c=TREE[node].child;
  while (TRUE) /* dummy loop, for "break" */
  {if (TREE[c].move==mv) break;
   else {c=TREE[c].sibling; if (!c) return -1;}}
  MakeMoveCheck(POS,mv,FALSE); node=TREE[c].loop?TREE[c].trans:c;}
 mv=get_move(POS,S); MakeMoveCheck(POS,mv,FALSE); *move=mv; return node;}

static void do_winloss(uint32 node,uint32 numbytes,char *ml,FILE *F)
{int c,mv; int parent=walk_to_pos(node,ml,&mv); fread(BYTE_ARR,1,numbytes,F);
 if (parent==-1) return;
 if (ml[0]) /* if no move list, everything is already ok */
 {c=TREE[parent].child; // ignore if not in book
  while (c) {if (TREE[c].move==mv) break; c=TREE[c].sibling;}}
 // else c=1; if (c) save_bytes(F,POS->DYN->HASH,numbytes);}
  else c=1; if (c) new_known(POS->DYN->HASH,BYTE_ARR,numbytes);}

static uint32 PAR=0;
static void install_result(uint32 node,uint32 good,uint32 bad,char *ml)
{int c,mv; int parent=walk_to_pos(node,ml,&mv); if (parent==-1) return;
 c=TREE[parent].child; // ignore if already in book, or if parent has result
 while (c) {if (TREE[c].move==mv) return; c=TREE[c].sibling;}
 if (!TREE[parent].good || !TREE[parent].bad) return;
 new_node(TREE[0].size,good,bad,parent,mv); TREE[0].size++; PAR=parent;}
// don't want to partially install children in some cases
// so if parent has result, ignore all

static void flush_parent()
{uint32 node; node=TREE[PAR].child; if (!node) return;
 if (BUG_TRACKING) {fprintf(BUGFILE,"FLUSH %d\n",PAR); fflush(BUGFILE);}
 do {if (!TREE[node].bad && TREE[node].flags&16 &&
	 !IsKnownResult(TREE[node].hash^WIN_TWIST)) win_installer(node);
     if (!TREE[node].good && !(TREE[node].flags&16) &&
	 !IsKnownResult(TREE[node].hash^LOSS_TWIST)) loss_installer(node);}
 while ((node=TREE[node].sibling));
 node_ratio_backtrack(PAR); backtrack_loopA(PAR);}

static void result(char *S)
{char ID[256],G[256],B[256]; int id,g,b; uint32 node,trans;
 sscanf(S,"%s %s %s",ID,G,B); id=atoi(ID); g=atoi(G); b=atoi(B);
 node=TABLE_NODE[id]; trans=TREE[node].loop?TREE[node].trans:node;
 TREE[trans].killer=0;
 install_result(trans,g,b,S+strlen(ID)+strlen(G)+strlen(B)+3);}

static void redo_result(uint32 n,uint32 g,uint32 b)
{TREE[n].good=g; TREE[n].bad=b; if (!TREE[n].size) TREE[n].size=TIME--;
 if (BUG_TRACKING) {fprintf(BUGFILE,"REDO %d %d %d\n",n,g,b); fflush(BUGFILE);}
 if (g==0) TREE[n].rat=0.0; else TREE[n].rat=FLOAT_INF; TREE[n].child=0;
 if (g==0) loss_installer(n); else win_installer(n);
 backtrack_loopA(TREE[n].parent);}

static void winloss(char *S,FILE *F) // size=TIME was already set
{char *U,ID[256],G[256],B[256],N[256]; int id,nb,g,b; uint32 node,trans;
 sscanf(S,"%s %s %s %s",ID,G,B,N); id=atoi(ID); nb=atoi(N);
 node=TABLE_NODE[id]; trans=TREE[node].loop?TREE[node].trans:node;
 if (BUG_TRACKING)
 {fprintf(BUGFILE,"winloss %d %d %d\n",id,node,trans); fflush(BUGFILE);}
 g=atoi(G); b=atoi(B); TREE[trans].killer=0;
 U=S+strlen(ID)+strlen(G)+strlen(B)+strlen(N)+4; do_winloss(trans,nb,U,F);
 if (!U[0]) redo_result(trans,g,b);}

static void quit_slave(int w)
{char B[32]="quit\n"; if (IS_DEAD[w]) return;
 fwrite(B,1,strlen(B),WRITE[w]); fflush(WRITE[w]);}

static void retire(char *S,int w)
{TABLE_STATUS[atoi(S)]=IS_DONE; WORKING--;
 if (IS_ALIVE[w]) NUM_AVAIL[w]++;
 else {MAX_AVAIL[w]--; if (NUM_AVAIL[w]==MAX_AVAIL[w]) quit_slave(w);}}

static void do_proc_slave_info(FILE *F,int w)
{char L[65536]; while (TRUE)
 {fgets(L,65000,F);
  if (BUG_TRACKING) {fprintf(BUGFILE,"FROM%d %s",w,L); fflush(BUGFILE);}
  if (!strcmp(L,"quitted\n"))
  {IS_DEAD[w]=TRUE; fclose(WRITE[w]); fclose(READ[w]); return;}
  if (!memcmp(L,"winloss ",8)) {winloss(L+8,F); continue;}
  if (!strcmp(L,"flush\n")) {flush_parent(); continue;}
  if (!memcmp(L,"retire ",7))
  {retire(L+7,w); if (LOGGING) {printf("%d< %s",w,L); fflush(stdout);}
   return;} // this is bad when there is more data!
  if (!memcmp(L,"result ",7)) {result(L+7); continue;}
  printf("Bad input from slave? %s\n",L);}}

////////////////////////////////////////////////////////////////////////

static char SPIN_ROOT[4096];
static int QUEUE_PTR=0,FIRST_IN_QUEUE=0;

static void ratio_sort_childrenA(uint32 node);

static uint64 RANDOM64()
{uint32 x=(uint32) random(),y=(uint32) random(),z=(uint32) random();
 x&=0x1fffff; y&=0x1fffff;
 return (((uint64) (x)) | (((uint64) (y))<<21) | (((uint64) (z))<<42));}

static void init_queue()
{int i; for (i=0;i<QUEUE_SIZE;i++) TABLE_STATUS[i]=TABLE_NODE[i]=0;
 QUEUE_PTR=0;}

static inline double WARPit(uint32 n,double z) // return 1.0/(x*x);
{double x=TREE[n].rat; // if (IsKnownLoss(TREE[n].hash^LOSS_TWIST)) x=x/25.0;
 if (x==FLOAT_INF) return 0.0; x=MIN(x,z*WIDE_LIMIT); // WIDE_LIMIT is 30.0
 return pow(x,-WIDE_POWER);} // WIDE_POWER is 2.0 // return 1.0/(x*x);

static uint32 random_child(uint32 n)
{uint32 c=TREE[n].child; double v,u=TREE[c].rat,total=WARPit(c,u);
 if (TREE[c].killer==IGNORE_IT) total=0;
 while ((c=TREE[c].sibling)) if (TREE[c].killer!=IGNORE_IT) total+=WARPit(c,u);
 v=((RANDOM64()&0xfffff)/1048576.0)*total;
 c=TREE[n].child; if (TREE[c].killer!=IGNORE_IT) v-=WARPit(c,u);
 while (v>=0.0 && (c=TREE[c].sibling))
   if (TREE[c].killer!=IGNORE_IT) v-=WARPit(c,u);
 if (!c) c=TREE[n].child; return c;}

static uint32 random_ratio_walk(uint32 node)
{uint16 mv; uint32 c,n; n=TREE[node].loop?TREE[node].trans:node;
 while (TREE[n].child) // MakeMove should be correct here, don't want reps
 {ratio_sort_childrenA(n); c=random_child(n); mv=TREE[c].move;
  sort_childrenA(n); MakeMove(POS,mv); n=TREE[c].loop?TREE[c].trans:c;}
 return n;}

static void get_string(int k,char *S)
{typeDYNAMIC *D=POS->DYN_ROOT+1; char N[64]; sprintf(S,"search %d",k);
 while (D!=POS->DYN) {strcat(S," "); strcat(S,Notate(N,(D+1)->mv)); D++;}}

static void inc_qptr()
{QUEUE_PTR++; if (QUEUE_PTR==QUEUE_SIZE) QUEUE_PTR=0;}

static boolean sent_ok(uint32 node,int k)
{char STR[4096]; if (TREE[node].loop) node=TREE[node].trans;
 if (TREE[node].killer) return FALSE;
 if (!TREE[node].bad || !TREE[node].good) return FALSE;
 if (TREE[node].child) return FALSE;
 TABLE_STATUS[QUEUE_PTR]=IS_WORKING; TABLE_NODE[QUEUE_PTR]=node;
 TREE[node].killer=IS_WORKING; get_string(QUEUE_PTR,STR); strcat(STR,"\n");
 if (LOGGING) {printf("%d> %s",k,STR); fflush(stdout);}
 if (BUG_TRACKING) {fprintf(BUGFILE,"TO%d %s",k,STR); fflush(BUGFILE);}
 fwrite(STR,1,strlen(STR),WRITE[k]); fflush(WRITE[k]);
 NUM_AVAIL[k]--; WORKING++; inc_qptr(); return TRUE;}

static boolean do_queue(char *A)
{boolean b; uint32 node,trans; int u; typeDYNAMIC *D; D=POS->DYN_ROOT+1;
 init_position(POS,"startpos"); parse_moves(POS,A);
 node=align_booktreeB(1,&b); if (!node) return FALSE; if (b) return FALSE;
 if (!TREE[node].bad || !TREE[node].good) return FALSE;
 trans=TREE[node].loop?TREE[node].trans:node;
 if (TREE[trans].killer) return FALSE; if (TREE[trans].child) return FALSE;
 if (LOGGING) {printf("Queue: %d %s\n",QUEUE_PTR,A); fflush(stdout);}
 TABLE_STATUS[QUEUE_PTR]=IN_QUEUE; TABLE_NODE[QUEUE_PTR]=node;
 u=0; while (D!=POS->DYN) {D++; TABLE_PATH[QUEUE_PTR][u++]=D->mv;}
 TABLE_PATH[QUEUE_PTR][u]=0;
 TREE[trans].killer=IN_QUEUE; QUEUED++; inc_qptr(); return TRUE;}
// check if 5 pc TB // check if a possible child is won

static boolean queue_to_work(int n,int k)
{boolean b; char STR[4096]; uint32 ck,node=TABLE_NODE[n],trans;
 if (TABLE_STATUS[n]==IS_DONE) return FALSE;
 trans=TREE[node].loop?TREE[node].trans:node; TABLE_STATUS[n]=IS_DONE;
 if (TREE[trans].killer==IS_WORKING) return FALSE;
 if (TREE[trans].killer==IN_QUEUE) QUEUED--; TREE[trans].killer=IS_DONE;
 if (!TREE[trans].bad || !TREE[trans].good) return FALSE;
 if (TREE[trans].child) return FALSE;
 init_position(POS,"startpos"); walk_path(n);
 ck=align_booktreeA(1,&b); if (b) return FALSE;
 if (ck!=trans) {printf("queue_to_work %d %d\n",ck,trans); return FALSE;}
 TABLE_STATUS[n]=IS_WORKING; TREE[trans].killer=IS_WORKING;
 get_string(n,STR); strcat(STR,"\n");
 if (LOGGING) {printf("%d> %s",k,STR); fflush(stdout);}
 if (BUG_TRACKING) {fprintf(BUGFILE,"TO%d %s",k,STR); fflush(BUGFILE);}
 fwrite(STR,1,strlen(STR),WRITE[k]); fflush(WRITE[k]);
 NUM_AVAIL[k]--; WORKING++; return TRUE;}

/* sent_ok() and queue_to_work() should ideally do TB checks, etc. */

static void SaveBook();
static boolean DONE=FALSE;
static void biff_spin_root()
{int n=strlen(SPIN_ROOT); if (DONE) return;
 if (n==0) {SaveBook(); DONE=TRUE; return;}
 if (n==4 || n==5) {SPIN_ROOT[0]=0; return;}
 if (SPIN_ROOT[n-4]==' ') SPIN_ROOT[n-4]=0;
 if (SPIN_ROOT[n-5]==' ') SPIN_ROOT[n-5]=0;
 if (BUG_TRACKING) {fprintf(BUGFILE,"ROOT %s\n",SPIN_ROOT); fflush(BUGFILE);}
 printf("New SPIN_ROOT %s\n",SPIN_ROOT); fflush(stdout);}

static void check_queue()
{uint32 node; boolean b; int w,i,c,j,ns=NUM_SLAVES;
 c=QUEUE_PTR; if (c<FIRST_IN_QUEUE) c+=QUEUE_SIZE;
 for (w=FIRST_IN_QUEUE;w<c;w++)
 {for (i=0;i<ns;i++) if (IS_ALIVE[i] && NUM_AVAIL[i]) break; if (i==ns) return;
  j=w&(QUEUE_SIZE-1); FIRST_IN_QUEUE=j; queue_to_work(j,i); NANO(1000000);}
 if (PAUSE) return;
 for (i=0;i<NUM_SLAVES;i++)
 {if (!IS_ALIVE[i]) continue; c=NUM_AVAIL[i]; for (j=0;j<c;j++)
  {init_position(POS,"startpos"); parse_moves(POS,SPIN_ROOT);
   node=align_booktreeA(1,&b); if (b || !node) {biff_spin_root(); continue;}
   node=random_ratio_walk(node); sent_ok(node,i); NANO(1000000);}}}

static void do_quit()
{int i; char B[32]="quit\n"; remove("KNOWN.lock");
 for (i=0;i<NUM_SLAVES;i++)
 {if (!IS_DEAD[i]) fwrite(B,1,strlen(B),WRITE[i]); fflush(WRITE[i]);}
 QUIT=TRUE;}

#include <errno.h>
static void do_register(char *A)
{int i,to_prog[2],from_prog[2]; pid_t p1;
 char PARAMS[256],COMMAND[1024],B[256]; char U[4096];
 sscanf(A,"%s",PARAMS); strcpy(COMMAND,A+strlen(PARAMS)+1);
 printf("REGISTER %s\n",COMMAND); pipe(to_prog); pipe(from_prog); p1=fork();
 if (!p1) // child
 {dup2(to_prog[0],STDIN_FILENO); dup2(from_prog[1],STDOUT_FILENO);
  close(to_prog[0]); close(to_prog[1]);
  close(from_prog[0]); close(from_prog[1]);
  execlp(COMMAND,COMMAND,NULL);
  printf("ERROR %s %d\n",COMMAND,errno); exit(1);}
 else {close(to_prog[0]); close(from_prog[1]);}
 fflush(stdout); WRITE[NUM_SLAVES]=fdopen(to_prog[1],"w");
 strcpy(B,A); for (i=0;i<strlen(B);i++) if (B[i]=='/') B[i]=' ';
 sscanf(B,"%d",&NUM_AVAIL[NUM_SLAVES]);
 MAX_AVAIL[NUM_SLAVES]=NUM_AVAIL[NUM_SLAVES];
 READ[NUM_SLAVES]=fdopen(from_prog[0],"r"); fd_in[NUM_SLAVES]=from_prog[0];
 setvbuf(READ[NUM_SLAVES],NULL,_IONBF,0);
 setvbuf(WRITE[NUM_SLAVES],NULL,_IONBF,0);
 while (TRUE)
 {fgets(U,4000,READ[NUM_SLAVES]);
  if (strstr(U,"ERROR")) {printf("%s\nFAILED %s\n",U,COMMAND); goto END;}
  if (!strcmp(U,"slaveready\n")) break;}
 sprintf(B,"params %s\n",PARAMS); fwrite(B,1,strlen(B),WRITE[NUM_SLAVES]);
 fflush(WRITE[NUM_SLAVES]); printf("REGISTERED %s\n",COMMAND);
 IS_DEAD[NUM_SLAVES]=FALSE; IS_ALIVE[NUM_SLAVES]=TRUE;
 strcpy(REGISTER_STRING[NUM_SLAVES],A); NUM_SLAVES++;
 END: printf("DONE WITH REGISTER\n"); fflush(stdout);}

////////////////////////////////////////////////////////////////////////

typedef struct {float rat; uint32 n;} RSORTER;
static int ratio_sort_compare(const void *x, const void *y)
{RSORTER A=*((RSORTER*) x); RSORTER B=*((RSORTER*) y);
 if (A.rat>B.rat) return 1; if (A.rat<B.rat) return -1; return 0;}

static void ratio_sort_childrenA(uint32 node)
{RSORTER S[256]; int i,u=0; uint32 n=TREE[node].child;
 S[u].rat=TREE[n].rat; S[u++].n=n;
 while (TREE[n].sibling)
 {n=TREE[n].sibling; S[u].rat=TREE[n].rat; S[u++].n=n;}
 qsort(S,u,sizeof(RSORTER),ratio_sort_compare);
 n=S[0].n; TREE[node].child=S[0].n;
 for (i=1;i<u;i++) {TREE[n].sibling=S[i].n; n=S[i].n;} TREE[n].sibling=0;}

static void spit_ratio(float x)
{float l=log(x); printf(" ");
 if (x==FLOAT_INF) printf("+INF  "); else if (x==0.0) printf("-INF  ");
 else if (l>=9.9995) printf("+%.2f",l);
 else if (l<=-9.9995) printf("%.2f",l);
 else if (l>=0.0) printf("+%.3f",l); else printf("%.3f",l); printf("  ");}

static int LIM;
#define GET_BIT(A,x) (A[(x)>>6]&(1ULL<<((x)&0x3f)))
#define SET_BIT(A,x) A[(x)>>6]|=(1ULL<<((x)&0x3f))
static void size_walk(uint32 n,uint64 *A) // trans is bogus?
{n=TREE[n].trans; if (GET_BIT(A,n)) return; LIM++; if (LIM>=1000000) return;
 SET_BIT(A,n); n=TREE[n].child; if (!n) return; size_walk(n,A);
 while ((n=TREE[n].sibling)) size_walk(n,A);}

static uint64 size_of_subtree(uint32 n) // dubious with trans
{uint64 *ARR; uint64 sz,u,ct=0; n=TREE[n].trans; if (!TREE[n].child) return 1;
 sz=(TREE[0].size+127)/64; ARR=malloc(sz*8+16384); LIM=0;
 for (u=0;u<sz;u++) ARR[u]=0; size_walk(n,ARR);
 for (u=0;u<sz;u++) ct+=POPCNT(ARR[u]); free(ARR); return ct;}

// static char* get_name(uint64 Z,char *A)
// {sprintf(A,"KNOWN/0x%02x/0x%016llx",(uint32)Z&0xff,Z); return A;}

// #include <sys/types.h>
// #include <sys/stat.h>
uint32 follow_known(typePOS *POS,typeDYNAMIC *D);

// static uint32 exists_known(uint64 Z)
// {char KNOWN_NAME[256]; struct stat buf;
// return (stat(get_name(Z,KNOWN_NAME),&buf)!=-1);}

#define IGNORE_WL TRUE
static void info_book_node(uint32 n,boolean b)
{uint32 u,k=TREE[n].child; char A[8]; boolean known; uint64 sz;
 printf("Val: "); spit_ratio(TREE[n].rat); spit(TREE+n);
 printf(" %lld",size_of_subtree(n));
 if (b) printf(" *"); printf("\n");
 while (k)
 {if (IGNORE_WL && ((TREE[TREE[n].child].bad && !TREE[k].bad) ||
		    (!TREE[TREE[n].child].good && TREE[k].good)))
  {k=TREE[k].sibling; continue;}
  printf("%s ", Notate(A,TREE[k].move));
  spit_ratio(TREE[k].rat); spit(TREE+k); u=TREE[k].loop?TREE[k].trans:k;
  sz=size_of_subtree(k); known=FALSE;
  //could also check that result is WON or LOST here...
  if (sz==1)
  {MakeMove(POS,TREE[k].move); if (known_find(POS->DYN->HASH)!=-1) known=TRUE;
   UnmakeMove(POS,TREE[k].move,TRUE);}
  printf(" %lld%s %d %d\n",sz,
	 (known || TREE[k].flags&4 || TREE[k].flags&64)?"*":"",
	 TREE[u].killer,TREE[k].size); k=TREE[k].sibling;}}

static void output_book(uint32 NODE,uint64 Z,boolean b)
{uint32 t=TREE[NODE].loop?TREE[NODE].trans:NODE;
 printf ("BOOK %d %d %d %llx\n",NODE,t,TREE[NODE].parent,Z);
 ratio_sort_childrenA(t); info_book_node(t,b); sort_childrenA(t);}

#if 0
static uint32 FindHash(uint64 HASH)
{int i; // printf("FindHash %llx\n",HASH);
 for (i=1;i<TREE[0].size;i++)
 {if (TREE[i].hash!=HASH) continue;
  if (TREE[i].flags&16 && TREE[i].good==MY_INFoo) return i;
  if ((TREE[i].flags&16)==0 &&  TREE[i].bad==MY_INFoo) return i;} return 0;}
#endif

static uint32 align_booktreeQ(uint32 NODE,boolean *b) // try follow TRANS W/L
{typeDYNAMIC *D; *b=FALSE; D=POS->DYN_ROOT+1;
 while (D!=POS->DYN && NODE)
 {D++; NODE=TreeMove(D->mv,NODE); // if (NODE==0) NODE=FindHash(D->HASH);
  if (!TREE[NODE].bad || !TREE[NODE].good) *b=TRUE;
  if (!NODE && known_find(D->HASH)!=-1) return follow_known(POS,D);
  if (TREE[NODE].loop || TREE[NODE].flags&4) NODE=TREE[NODE].trans;
  if (!TREE[NODE].child && known_find(D->HASH)!=-1)
   return follow_known(POS,D);}
 return NODE;}

#define SAVE_GAP 100000
static uint32 LAST_SAVE;
static void check_save() {if (TREE[0].size>LAST_SAVE+SAVE_GAP) SaveBook();}

static int PASCNT=0;
static void doit_query()
{uint32 t,NODE; boolean b; PASCNT++;
 if ((PASCNT%16)==0)
   printf("Book size %d/%d\n",TREE[0].size,TREE[0].size-LAST_SAVE);
 fflush(stdout); NODE=align_booktreeQ(1,&b);
 if (NODE==0xffffffff) return; // HACK for follow_known
 t=TREE[NODE].loop?TREE[NODE].trans:NODE;
 if (!NODE) printf("Not in BOOK\n");
 else if (!TREE[t].child) printf("No children %d 0x%llx %d/%d\n",
				 t,POS->DYN->HASH,TREE[t].good,TREE[t].bad);
 else output_book(NODE,POS->DYN->HASH,b); printf("Book Query Complete\n");
 printf("status %d %d %.2f %.1f\n",QUEUED,WORKING,WIDE_POWER,WIDE_LIMIT);
 fflush(stdout);}

static void do_getinfo(char *A)
{init_position(POS,"startpos"); parse_moves(POS,A); doit_query();
 if (known_find(POS->DYN->HASH)!=-1)
 {boolean b; uint32 t; t=align_booktreeA(1,&b);
  if (t && TREE[t].good && TREE[t].bad && !TREE[t].child)
  {POS->wtm?win_installer(t):loss_installer(t); // should even install w/child
   backtrack_loopA(TREE[t].parent); node_ratio_backtrack(TREE[t].parent);}}}

static void do_ignore(char *A)
{init_position(POS,"startpos"); if (!parse_moves(POS,A)) return;
 boolean b; uint32 n=align_booktreeA(1,&b); if (!n) return; 
 if (TREE[n].killer==0) TREE[n].killer=3;
 else if (TREE[n].killer==3) TREE[n].killer=0;}

static void do_new_root(char *A)
{init_position(POS,"startpos");
 if (!parse_moves(POS,A)) return; strcpy(SPIN_ROOT,A);}

static void clear_queue()
{int i;
 for (i=0;i<QUEUE_SIZE;i++)
  if (TABLE_STATUS[i]==IN_QUEUE) TABLE_STATUS[i]=IS_DONE;
 FIRST_IN_QUEUE=QUEUE_PTR;}

static void do_bye_slave(int w)
{IS_ALIVE[w]=0; if (MAX_AVAIL[w]==NUM_AVAIL[w]) quit_slave(w);}

static void list_slaves()
{int i; printf("SLAVE LIST\n"); fflush(stdout);
 for (i=0;i<NUM_SLAVES;i++)
 {if (IS_DEAD[i]) continue;
  printf("Slave %d%s: %s ",i,IS_ALIVE[i]?" ":" [DYING]",REGISTER_STRING[i]);
  printf("%dw/%di\n",MAX_AVAIL[i]-NUM_AVAIL[i],NUM_AVAIL[i]);}
 printf("status %d %d %.2f %.1f\n",QUEUED,WORKING,WIDE_POWER,WIDE_LIMIT);
 fflush(stdout);}

static void do_wide_power(double x)
{if (x<1.0) x=1.0; if (x>3.0) x=3.0; WIDE_POWER=x;
 if (0) printf("Changed WIDE_POWER to %f\n",WIDE_POWER); fflush(stdout);}

static void do_wide_limit(double x)
{if (x<1.0) x=1.0; if (x>1000.0) x=1000.0; WIDE_LIMIT=x;
 if (0) printf("Changed WIDE_LIMIT to %f\n",WIDE_LIMIT); fflush(stdout);}

static void do_input()
{char A[4096]; fgets(A,4000,stdin); A[strlen(A)-1]=0;
 if (BUG_TRACKING) {fprintf(BUGFILE,"INPUT %s\n",A); fflush(BUGFILE);}
 if (!memcmp(A,"register ",9)) do_register(A+9);
 if (!memcmp(A,"root ",5)) do_new_root(A+5); // location of background spin
 if (!strcmp(A,"clearqueue")) clear_queue(); // QUEUE_PTR
 if (!memcmp(A,"byeslave",8)) do_bye_slave(atoi(A+9));
 if (!memcmp(A,"queue ",6)) do_queue(A+6);
 if (!memcmp(A,"ignore ",7)) do_ignore(A+7);
 if (!memcmp(A,"widepower",9)) do_wide_power(atof(A+10));
 if (!memcmp(A,"widelimit",9)) do_wide_limit(atof(A+10));
 if (!strcmp(A,"listslaves")) list_slaves();
 if (!memcmp(A,"getinfo ",8)) do_getinfo(A+8); // same as "query"
 if (!strcmp(A,"pause")) PAUSE=TRUE; if (!strcmp(A,"resume")) PAUSE=FALSE;
 if (!strcmp(A,"savebook")) SaveBook();
 if (!strcmp(A,"quit")) do_quit(); fflush(stdout);}

static void poll_slaves()
{fd_set fd[1]; int i; struct timeval tv[1]; tv->tv_sec=tv->tv_usec=0;
 for (i=0;i<NUM_SLAVES;i++)
 {if (IS_DEAD[i]) continue; FD_ZERO(fd); FD_SET(fd_in[i],fd);
  if (select(fd_in[i]+1,fd,NULL,NULL,tv)<=0) continue;
  do_proc_slave_info(READ[i],i);}}

static void master_loop()
{while (!QUIT)
 {poll_slaves(); while (TryInput()) do_input();
  check_queue(); check_save(); NANO(10000000);}}

////////////////////////////////////////////////////////////////////////

static void LoadNode(FILE *F,PN_NODE *P,uint32 w)
{fread(&(P->good),4,1,F); fread(&(P->bad),4,1,F); fread(&(P->size),4,1,F);
 fread(&P->hash,8,1,F); fread(&(P->trans),4,1,F); fread(&(P->loop),4,1,F);
 fread(&(P->parent),4,1,F); fread(&(P->sibling),4,1,F);
 fread(&(P->child),4,1,F); fread(&(P->move),2,1,F);
 fread(&(P->killer),1,1,F); fread(&(P->flags),1,1,F); P->killer=0; // ok
 if (P->child || (P->loop && P->trans!=w)) P->rat=-1.0;
 else {if (!(P->bad)) P->rat=FLOAT_INF; else P->rat=(TEMPO*P->good)/P->bad;}}

static int SaveNode(FILE *F,PN_NODE *P)
{int u=0; u+=fwrite(&(P->good),4,1,F); u+=fwrite(&(P->bad),4,1,F);
 u+=fwrite(&(P->size),4,1,F); u+=fwrite(&(P->hash),8,1,F);
 u+=fwrite(&(P->trans),4,1,F); u+=fwrite(&(P->loop),4,1,F);
 u+=fwrite(&(P->parent),4,1,F); u+=fwrite(&(P->sibling),4,1,F);
 u+=fwrite(&(P->child),4,1,F); u+=fwrite(&(P->move),2,1,F);
 u+=fwrite(&(P->killer),1,1,F); u+=fwrite(&(P->flags),1,1,F); return u;}

static char BOOK_PATH[1024];
static void SaveBook()
{FILE *F; uint32 u; char COPY_PATH[1024];
 LAST_SAVE=TREE[0].size; printf("Book size %d\n",TREE[0].size);
 printf("SAVING BOOK %s %d pos\n",BOOK_PATH,TREE[0].size); fflush(stdout);
 strcpy(COPY_PATH,BOOK_PATH); strcat(COPY_PATH,".tmp"); F=fopen(COPY_PATH,"w");
 if (!F) {printf("BOOK SAVE %s FAILED!!\n",BOOK_PATH); fflush(stdout); return;}
 for (u=0;u<TREE[0].size;u++) if (SaveNode(F,TREE+u)!=12)
 {printf("BOOK SAVE FAILED!! %d\n",u); fflush(stdout); fclose(F); return;}
 fclose(F); rename(COPY_PATH,BOOK_PATH); // rename .tmp to orig name
 printf("BOOK SAVE %s saved %d pos\n",BOOK_PATH,LAST_SAVE); fflush(stdout);}

/************************************************************************/

static void load_book()
{PN_NODE P[1]; uint32 u,sz,all; FILE* F; char A[64];
 F=fopen(BOOK_PATH,"r");
 if (!F) {strcpy(A,"BOOKS/"); strcat(A,BOOK_PATH);
          F=fopen(A,"r"); if (F) strcpy(BOOK_PATH,A);}
 if (!F) {printf("%s not found\n",BOOK_PATH); exit(1);}
 LoadNode(F,P,0); sz=P->size; printf("BOOK %s has %d pos ",BOOK_PATH,sz);
 printf("Book size %d\n",sz); fflush(stdout);
 all=2<<BSR((8192+sz*3)/2); if (all<(1<<24)) all=(1<<24);
 // printf("allocating %d\n",all);
 TREE=CALLOC(all,sizeof(PN_NODE)); TREE[0].size=sz;
 for (u=1;u<sz;u++) LoadNode(F,TREE+u,u); fclose(F); BOOK_HASH_MASK=all-1;
 BOOK_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 all=2<<(BSR(sz+sz/2+0x100000)); all=1024; ALL_HASH_MASK=all-1;
 ALL_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) ALL_HASH[u].key=0;
 init_position(POS,"startpos");
 InitBookHash(1); initial_book_scan(); init_all_ratios();
 for (u=0;u<sz;u++) TREE[u].killer=0; TIME=0xffffffffULL;
 for (u=1;u<sz;u++) if (TREE[u].size &&TREE[u].size<=TIME) TIME=TREE[u].size-1;
 LAST_SAVE=TREE[0].size; printf("Book size %d\n",TREE[0].size);
 printf("BOOK LOADED TIME 0x%x\n",(uint32) TIME); fflush(stdout);}

////////////////////////////////////////////////////////////////////////

void simple_tb_setup();
static void init_stuff(int c,char **I)
{BOOK_PATH[0]=0; if (c>1) strcpy(BOOK_PATH,I[1]); magic_mult_init();
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); SPIN_ROOT[0]=0;
 POS->DYN=POS->DYN_ROOT+1; srandom((uint32) Now()); BYTE_ARR=malloc(1<<25);
 init_position(POS,"startpos"); printf("LOSING CHESS\n"); fflush(stdout);
 char BUGNAME[1024]; sprintf(BUGNAME,"/tmp/LOSING.dump.%d",getpid());
 if (BUG_TRACKING) BUGFILE=fopen(BUGNAME,"w");
 simple_tb_setup(); PAUSE=TRUE; init_queue(); load_book(); // exterior_wins();
 init_known(TRUE); setvbuf(stdin,NULL,_IONBF,0); setvbuf(stdout,NULL,_IONBF,0);
 printf("CLUSTER MASTER\n"); fflush(stdout);}

int main(int c,char **I)
{init_stuff(c,I); master_loop(); return 0;}
