
#include "losing.h"

// bit 31 of data set, has trans
// bit 30 of data set, has child, or has sibling when has trans
// 10 trans, no sibling, data is trans
// 01 has child (+1), data is sibling (or zero)
// 11 trans and sibling (+1), data is trans

#define DATA_IS_TRANS(A,x) (((A[x].data&(1U<<31)))==(1U<<31))
#define DATA_HAS_CHILD(A,x) (((A[x].data&(3U<<30)))==(1U<<30))
#define TRANS_AND_SIBLING(A,x) (((A[x].data&(3U<<30)))==(3U<<30))
#define DATA_GET_NODE(A,x) (A[x].data&0x3fffffff)
#define SET_CHILD(A,x) (A[x].data|=(1U<<30))
#define NEXT_SIBLING(A,x) ((TRANS_AND_SIBLING(A,x)) ? ((x)+1) : \
			   (DATA_IS_TRANS(A,x) ? 0 : DATA_GET_NODE(A,x)))

static tTRANS *TRANS;
static uint64 TRANS_MASK;
static tTRANS *NODE_TABLE;
static uint32 *REDIRECT;

static boolean is_unsolved(uint16 mv)
{if ((mv!=0xfedc) && ((mv>>12)>8 || (mv>>12)==7)) return TRUE; return FALSE;}

static uint32 LookupRehash(typePOS *POS,uint32 n)
{uint64 u,Z=POS->DYN->HASH; uint32 k=Z&TRANS_MASK; if (!k) k++;
 while ((u=TRANS[k].key))
 {if (u==Z) return TRANS[k].node; k++; if (k==(TRANS_MASK+1)) k=1;}
 TRANS[k].key=Z; TRANS[k].node=n; return 0;}

static uint32 RehashStep
(typePOS *POS,WIN_NODE *TREE,uint32 NODE,WIN_NODE *WIN,uint32 wn)
{uint32 u,c,x,last; uint16 mv;
 SET_CHILD(WIN,wn); wn++; last=wn; c=NODE+1; mv=TREE[c].move;
 WIN[wn].move=mv; WIN[wn].data=0; if (is_unsolved(mv)) {wn++; goto LOOP;}
 MakeMoveCheck(POS,mv,FALSE);
 if (DATA_IS_TRANS(TREE,c)) x=DATA_GET_NODE(TREE,c); else x=c; // wintree?
 while (x!=REDIRECT[x]) x=REDIRECT[x];
 if ((u=LookupRehash(POS,wn))) WIN[wn++].data=(1U<<31)|u;
 else if (DATA_HAS_CHILD(TREE,x)) wn=RehashStep(POS,TREE,x,WIN,wn); else wn++;
 UnmakeMove(POS,mv,TRUE);
 LOOP: while ((c=NEXT_SIBLING(TREE,c)))
 {mv=TREE[c].move; WIN[wn].move=mv; WIN[wn].data=0;
  if (DATA_IS_TRANS(WIN,last)) WIN[last].data|=(1U<<30);
  else WIN[last].data|=wn; last=wn;
  if (is_unsolved(mv)) {wn++; continue;} MakeMoveCheck(POS,mv,FALSE);
  if (DATA_IS_TRANS(TREE,c)) x=DATA_GET_NODE(TREE,c); else x=c; // wintree?
  while (x!=REDIRECT[x]) x=REDIRECT[x];
  if ((u=LookupRehash(POS,wn))) WIN[wn++].data=(1U<<31)|u;
  else if (DATA_HAS_CHILD(TREE,x)) wn=RehashStep(POS,TREE,x,WIN,wn); else wn++;
  UnmakeMove(POS,mv,TRUE);} return wn;}

// MakeMoveCheck can occasionally fail with long MAXREV
// only solution I know is to temporarily invrease in gen_moves.c

////////////////////////////////////////////////////////////////////////

static void NODE_TO_HASH(uint64 Z,boolean wtm,uint32 n)
{uint64 u; uint32 k=Z&TRANS_MASK;
 if (!wtm || NODE_TABLE[n].key) return; if (!k) k++; NODE_TABLE[n].key=Z;
 while ((u=TRANS[k].key))
 {if (u==Z) {NODE_TABLE[n].node=TRANS[k].node; return;}
  k++; if (k==(TRANS_MASK+1)) k=1;}
 TRANS[k].key=Z; TRANS[k].node=n; NODE_TABLE[n].node=n;}

static void fill_trans(typePOS *POS,WIN_NODE *WIN,uint32 n)
{uint32 c; uint16 mv; if (DATA_IS_TRANS(WIN,n)) n=DATA_GET_NODE(WIN,n);
 if (!DATA_HAS_CHILD(WIN,n) || NODE_TABLE[n].node) return;
 NODE_TO_HASH(POS->DYN->HASH,POS->wtm,n);
 c=n+1; mv=WIN[c].move; if (is_unsolved(mv)) goto LOOP;
 MakeMoveCheck(POS,mv,FALSE); fill_trans(POS,WIN,c); UnmakeMove(POS,mv,TRUE);
 LOOP: while ((c=NEXT_SIBLING(WIN,c)))
 {mv=WIN[c].move; if (is_unsolved(mv)) continue; MakeMoveCheck(POS,mv,FALSE);
  fill_trans(POS,WIN,c); UnmakeMove(POS,mv,TRUE);}}

static uint64 *ARR; static uint32 *SEEN; static uint32 num;
#define GET_ARRAY_BIT(A,x) (A[(x)>>6]&(1ULL<<((x)&0x3f)))
#define SET_ARRAY_BIT(A,x) {A[(x)>>6]|=(1ULL<<((x)&0x3f)); SEEN[num++]=x;}
#define ERASE_ARRAY_BIT(A,x) A[(x)>>6]&=~(1ULL<<((x)&0x3f))
//#define ERASE_ARRAY_BIT(A,x) A[(x)>>6]=0
static void walk_tree(WIN_NODE *W,uint32 n,uint32 t)
{uint32 c; if (t) {if (GET_ARRAY_BIT(ARR,n)) return; SET_ARRAY_BIT(ARR,n);}
 if (DATA_IS_TRANS(W,n)) {walk_tree(W,DATA_GET_NODE(W,n),t); return;}
 if (!t) {if (GET_ARRAY_BIT(ARR,n)) return; SET_ARRAY_BIT(ARR,n);}
 if (!DATA_HAS_CHILD(W,n)) return; c=n+1; walk_tree(W,c,t);
 while ((c=NEXT_SIBLING(W,c))) walk_tree(W,c,t);}

static uint32 compute_subtree_size(WIN_NODE *W,uint32 n)
{int i; if (!n) return W[0].data&0x3fffffff;
 while (DATA_IS_TRANS(W,n)) n=DATA_GET_NODE(W,n);
 num=0; walk_tree(W,n,TRUE); for (i=0;i<num;i++) ERASE_ARRAY_BIT(ARR,SEEN[i]);
 return num;}

static void CHOOSE(WIN_NODE *WIN,uint32 n,uint32 u)
{uint32 vn,vu; uint16 m1,m2; char N1[8],N2[8];
 m1=WIN[u+1].move; m2=WIN[n+1].move;
 vu=compute_subtree_size(WIN,u); vn=compute_subtree_size(WIN,n);
 if (0) printf("%d %d %s %s %d %d\n",u,n,Notate(N1,m1),Notate(N2,m2),vu,vn);
 if (vn<vu) REDIRECT[u]=REDIRECT[n]; else REDIRECT[n]=REDIRECT[u];}
// this is not quite always working, really want to see if a BIT is known
// to lead to a *result* in walk_tree(), not just take the smaller subgraph
// but even that is not foolproof, as the different subgraphs can conflict

// better (probably will never get implemented), run "verify" on new wintree,
// if it has loops, look in NODE_TABLE and change relevant REDIRECT
// then rebuild, iterate until done (no idea if it really works either)

static void scan_trans(WIN_NODE *WIN)
{uint32 n,u,v,sz=WIN[0].data&0x3fffffff; uint16 m1,m2;
 for (u=0;u<sz;u++)
 {if (u&&(u%100000)==0) {printf("%c",((u%1000000)==0)?'|':'+');fflush(stdout);}
  if (NODE_TABLE[u].key && NODE_TABLE[u].node!=u)
  {n=NODE_TABLE[u].node; v=u;
   while (DATA_IS_TRANS(WIN,n)) n=DATA_GET_NODE(WIN,n);
   while (DATA_IS_TRANS(WIN,v)) v=DATA_GET_NODE(WIN,v);
   if (!DATA_HAS_CHILD(WIN,v) || !DATA_HAS_CHILD(WIN,n)) continue;
   m1=WIN[v+1].move; m2=WIN[n+1].move; if (m1!=m2) CHOOSE(WIN,n,v);}}}

void newrehash(WIN_NODE *WIN,uint16 *ML,int w,char *fn,boolean SAVE)
{WIN_NODE *COPY; uint32 u; typePOS POS[1]; uint64 sz,sz2,alloc_size;
  // SaveWinTree(WIN,ML,w,fn); return;
 memset((void*) POS,0,sizeof(typePOS));
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 init_position(POS,"startpos"); printf ("REHASH WIN-TREE\n");
 sz=WIN[0].data&0x3fffffff; sz2=sz+sz/8+0x10000;
 TRANS_MASK=2<<BSR(sz2); printf("Input size is %lld records\n",sz);
 alloc_size=((uint64) TRANS_MASK)*sizeof(tTRANS); TRANS=malloc(alloc_size);
 printf("TRANS table at %lld entries [%lldmb]\n",TRANS_MASK,alloc_size>>20);
 for (u=0;u<TRANS_MASK;u++) TRANS[u].key=TRANS[u].node=0; TRANS_MASK--;
 NODE_TABLE=malloc((sz+256)*sizeof(tTRANS));
 for (u=0;u<sz;u++) NODE_TABLE[u].key=0;
 REDIRECT=malloc((sz+256)*sizeof(uint32)); for (u=0;u<sz;u++) REDIRECT[u]=u;
 for (u=0;u<w;u++) MakeMove(POS,ML[u]); // prolog
 ARR=malloc((sz/8+0x10000)); for (u=0;u<sz/64+0x1000;u++) ARR[u]=0;
 SEEN=malloc((sz+256)*sizeof(uint32)); fill_trans(POS,WIN,0);
 printf("Trans table filled\nScanning: "); scan_trans(WIN); printf("\n");
 free(NODE_TABLE); free(ARR); TRANS_MASK=2<<BSR(sz2);
 for (u=0;u<TRANS_MASK;u++) TRANS[u].key=TRANS[u].node=0; TRANS_MASK--;
 printf("Final rehash\n");
 sz=sz+sz/8+0x10000; COPY=CALLOC(sz,sizeof(WIN_NODE));
 for (u=0;u<sz;u++) COPY[u].data=COPY[u].move=0;
 COPY[0].data=RehashStep(POS,WIN,0,COPY,0); SET_CHILD(COPY,0);
 printf("Reduced to %d entries\n",COPY[0].data&0x3fffffff);
 if (SAVE) SaveWinTree(COPY,ML,w,fn); free(COPY); free(TRANS);
 free(REDIRECT); free(SEEN); free(POS->DYN_ROOT);}

// Need to be more careful in malloc/free

////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////

static uint32 try=0,copied=0,US=0;
static uint32 *LEV2_CHANGE;
static WIN_NODE *C; static uint64 SIZE=0; // global, for resizing

static void resize()
{uint64 ns=SIZE+SIZE/4; int u; printf("Resize wintree %lld to %lld\n",SIZE,ns);
 C=realloc(C,ns*sizeof(WIN_NODE));
 for (u=SIZE;u<ns;u++) C[u].data=C[u].move=0; SIZE=ns;}

static uint32 CopyWinTree(WIN_NODE *U,uint32 n,uint32 c) // C is global
{uint32 last=c; if (c>SIZE-256) resize();
 LEV2_CHANGE[n]=c; C[c].move=U[n].move;
 if (DATA_IS_TRANS(U,n)) C[c].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)];
 if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWinTree(U,n+1,++c);} else c++;
 while ((n=NEXT_SIBLING(U,n)))
 {LEV2_CHANGE[n]=c; C[c].move=U[n].move;
  if (DATA_IS_TRANS(C,last)) C[last].data|=(1U<<30); // TRANS_AND_SIBLING
  else C[last].data|=c; last=c;
  if (DATA_IS_TRANS(U,n))
  {C[c++].data=(1U<<31)|LEV2_CHANGE[DATA_GET_NODE(U,n)]; continue;}
  if (DATA_HAS_CHILD(U,n)) {SET_CHILD(C,c); c=CopyWinTree(U,n+1,++c);}
  else c++;} return c;}
// CopyWinTree actually assumes things are OK with trans in first tree
// (if DATA_IS_TRANS, then it has already been written to LEV2_CHANGE)
// so something like "subtree" needs a bit more work

static uint32 *NAME_CHANGE; // input size limited to 2^24?

static boolean is_terminal_white_wins(typePOS *POS)
{uint16 ML[256]; if (POPCNT(wBitboardOcc|bBitboardOcc)<=4) return TRUE;
 if (!POS->wtm) return FALSE; return (GenMoves(POS,ML)==0);}

static uint32 do_expand(typePOS *POS,WIN_NODE *W,uint32 wn,uint32 cn)
{uint64 KN,Z; uint16 mv; uint32 last=cn; if (cn>SIZE-256) resize();
 while (DATA_IS_TRANS(W,wn)) // should never happen
 {printf("TRANS %d %d\n",wn,DATA_GET_NODE(W,wn)); wn=DATA_GET_NODE(W,wn);}
 NAME_CHANGE[wn]=cn; try++; Z=POS->DYN->HASH; KN=known_find(Z);
 if (KN!=-1 || (wn && !DATA_HAS_CHILD(W,wn))) // what if mv tweaked?
 {WIN_STRUCT *WS; WIN_NODE *U; uint32 d;
  if (KN==-1) // issues move tweak
  {if (is_terminal_white_wins(POS)) return cn+1;
   printf("UNSOLVED %d 0x%016llx\n",wn,Z); US++;
   C[cn].move^=0xf000; return cn+1;}
  WS=WintreeKnown(Z); U=WS->W; d=U[0].data&0x3fffffff;
  if (d<=1) printf("Null copy %d to %d? 0x%llx\n",wn,cn,Z);
  if (d<=1) {free(U); free(WS->MOVE_LIST); free(WS); return cn+1;}
  if (1) printf("Copy %d to %d: %d records from 0x%llx\n",wn,cn,d-1,Z);
  SET_CHILD(C,cn); cn=CopyWinTree(U,1,cn+1); copied++;
  fflush(stdout); free(U); free(WS->MOVE_LIST); free(WS); return cn;}
 if (!wn || DATA_HAS_CHILD(W,wn)) // recurse to child
 {wn++; mv=W[wn].move; SET_CHILD(C,cn); cn++; C[cn].move=mv; last=cn;
  if (is_unsolved(mv)) {C[cn++].data=0; goto LOOP;}
  if (DATA_IS_TRANS(W,wn) && NAME_CHANGE[DATA_GET_NODE(W,wn)])
  {if (!TRANS_AND_SIBLING(W,wn))
   {C[cn++].data=(1U<<31)|NAME_CHANGE[DATA_GET_NODE(W,wn)]; return cn;}
   else C[cn++].data=(3U<<30)|NAME_CHANGE[DATA_GET_NODE(W,wn)];}
  else {MakeMove(POS,mv); cn=do_expand(POS,W,wn,cn); UnmakeMove(POS,mv,1);}
  // well, I'm not sure why MakeMove always works, but I've never seen a rep...
 LOOP:
  while ((wn=NEXT_SIBLING(W,wn)))
  {mv=W[wn].move; C[cn].move=mv;
   if (!DATA_IS_TRANS(C,last)) C[last].data|=cn; last=cn;
   if (is_unsolved(mv)) {C[cn++].data=0; continue;}
   if (DATA_IS_TRANS(W,wn) && NAME_CHANGE[DATA_GET_NODE(W,wn)])
   {if (!TRANS_AND_SIBLING(W,wn))
    {C[cn++].data=(1U<<31)|NAME_CHANGE[DATA_GET_NODE(W,wn)]; return cn;}
    else C[cn++].data=(3U<<30)|NAME_CHANGE[DATA_GET_NODE(W,wn)]; continue;}
   MakeMove(POS,mv); cn=do_expand(POS,W,wn,cn); UnmakeMove(POS,mv,1);}}
 return cn;}

static uint16 PROLOG[1024];

void expand_and_rehash(WIN_STRUCT *WS,char *fout,boolean SAVE)
{WIN_NODE *W; uint32 u,w,wsz; typePOS POS[1]; REV_LIMIT=125;
 memset((void*) POS,0,sizeof(typePOS));
 POS->DYN_ROOT=CALLOC(MAX_PLY,sizeof(typeDYNAMIC)); POS->DYN=POS->DYN_ROOT+1;
 init_position(POS,"startpos"); printf("EXPAND WINTREE\n"); fflush(stdout);
 W=WS->W; SET_CHILD(W,0); w=W[0].move;
 for (u=0;u<w;u++) PROLOG[u]=WS->MOVE_LIST[u];
 for (u=0;u<w;u++) MakeMove(POS,PROLOG[u]);
 NAME_CHANGE=malloc((1<<24)*sizeof(uint32)); // (1<<24) max size of input WT
 for (u=0;u<(1<<24);u++) NAME_CHANGE[u]=0;
 LEV2_CHANGE=malloc(0x1000000*sizeof(uint32)); // largest KNOWN proof (1<<24)
 wsz=1024+(W[0].data&0x3fffffff); SIZE=(1<<20)+100*wsz;
 if (SIZE>(1<<30)) SIZE=(1<<30);
 printf("Allocating %lldK records for COPY\n",SIZE>>10);
 C=malloc(SIZE*sizeof(WIN_NODE)); for(u=0;u<SIZE;u++) C[u].data=C[u].move=0;
 uint32 n; n=do_expand(POS,W,0,0); C[0].data=n|(1U<<30); // SPA, C can move
 printf("Tried %d pn2-nodes, copied %d, unsolved %d\n",try,copied,US);
 printf("Proof size is %d\n",C[0].data&0x3fffffff);
 free(POS->DYN_ROOT); free(NAME_CHANGE); free(LEV2_CHANGE);
 newrehash(C,PROLOG,w,fout,SAVE); free(C); REV_LIMIT=95;}
