
#include "losing.h"

// bit 31 of data set, has trans
// bit 30 of data set, has child, or has sibling when has trans
// 10 trans, no sibling, data is trans
// 01 has child (+1), data is sibling (or zero)
// 11 trans and sibling (+1), data is trans

#define DATA_IS_TRANS(A,x) (((A[x].data&(1U<<31)))==(1U<<31))
#define DATA_HAS_CHILD(A,x) (((A[x].data&(3U<<30)))==(1U<<30))
#define TRANS_AND_SIBLING(A,x) (((A[x].data&(3U<<30)))==(3U<<30))
#define DATA_GET_NODE(A,x) (A[x].data&0x3fffffff)
#define SET_CHILD(A,x) (A[x].data|=(1U<<30))
#define NEXT_SIBLING(A,x) ((TRANS_AND_SIBLING(A,x)) ? ((x)+1) : \
			   (DATA_IS_TRANS(A,x) ? 0 : DATA_GET_NODE(A,x)))

static int comp(const void *x,const void *y)
{uint16 a=*((uint16*) x); uint16 b=*((uint16*) y);
 if (a>b) return 1; if (a<b) return -1; return 0;}

static boolean is_unsolved(uint16 mv)
{if ((mv!=0xfedc) && ((mv>>12)>8 || (mv>>12)==7)) return TRUE; return FALSE;}

static void GET_JOINT_DESCENDANT_LIST // should ignore mv^0xf000
(uint16 *ML,WIN_NODE *W1,uint32 n1,WIN_NODE *W2,uint32 n2)
{int w=0,y=1; ML[0]=0; ML[1]=0;
 if (n1!=-1 && DATA_HAS_CHILD(W1,n1))
 {n1++; if (!is_unsolved(W1[n1].move)) ML[w++]=W1[n1].move;
  while ((n1=NEXT_SIBLING(W1,n1)))
  {if (!is_unsolved(W1[n1].move)) ML[w++]=W1[n1].move;}}
 if (n2!=-1 && DATA_HAS_CHILD(W2,n2))
 {n2++; if (!is_unsolved(W2[n2].move)) ML[w++]=W2[n2].move;
  while ((n2=NEXT_SIBLING(W2,n2)))
  {if (!is_unsolved(W2[n2].move)) ML[w++]=W2[n2].move;}}
 ML[w]=0; qsort(ML,w,sizeof(uint16),comp);
 for (w=1;ML[w];w++) if (ML[w]!=ML[w-1]) ML[y++]=ML[w]; ML[y]=0;}

typedef struct {uint64 key; uint32 main,t1,t2;} tTRANS3;
static tTRANS3* HASH_TABLE;
#define HASH_MASKING (HASH_LIMIT-1)
uint32 HASH_LIMIT;

static uint32 HASH_LOOKUP(uint64 Z)
{uint64 h; uint32 k=Z&HASH_MASKING; if (!k) k++;
 while ((h=HASH_TABLE[k].key))
 {if (h==Z) return k; k++; if (k==HASH_LIMIT) k=1;} return 0;}

static uint32 INCLUDE_HASH(uint64 Z,uint32 n,boolean w)
{uint64 h; uint32 k=Z&HASH_MASKING; if (!k) k++;
 while ((h=HASH_TABLE[k].key))
 {if (h!=Z) {k++; if (k==HASH_LIMIT) k=1; continue;}
  if (!w && HASH_TABLE[k].t1!=-1) return TRUE;
  if (w && HASH_TABLE[k].t2!=-1) return TRUE; break;}
 HASH_TABLE[k].key=Z;
 if (!w) HASH_TABLE[k].t1=n; else HASH_TABLE[k].t2=n; return FALSE;}

static void fill_hash_table(typePOS *POS,WIN_NODE *W,uint32 n,boolean w)
{uint16 mv; // this requires that HASH is tree-unique [can be done via rehash]
 if (INCLUDE_HASH(POS->DYN->HASH,n,w) || !DATA_HAS_CHILD(W,n)) return;
 n++; mv=W[n].move; if (is_unsolved(mv)) goto LOOP;
 MakeMoveCheck(POS,mv,FALSE); fill_hash_table(POS,W,n,w); UnmakeMove(POS,mv,0);
 LOOP: while ((n=NEXT_SIBLING(W,n)))
 {mv=W[n].move; if (is_unsolved(mv)) continue; MakeMoveCheck(POS,mv,FALSE);
  fill_hash_table(POS,W,n,w); UnmakeMove(POS,mv,0);} return;}

static uint32 StepWinTree
(typePOS *POS,WIN_NODE *W1,WIN_NODE *W2,WIN_NODE *NEW,uint32 NODE)
{uint32 h,n1,n2,LAST; uint16 G[256],mv; uint32 w=0; G[1]=0;
 h=HASH_LOOKUP(POS->DYN->HASH); if (!h) printf("ERROR\n");
 if (HASH_TABLE[h].main)
 {NEW[NODE++].data=(1U<<31)|HASH_TABLE[h].main; return NODE;}
 HASH_TABLE[h].main=NODE; n1=HASH_TABLE[h].t1; n2=HASH_TABLE[h].t2;
 if ((n1==-1 || !DATA_HAS_CHILD(W1,n1)) &&
     (n2==-1 || !DATA_HAS_CHILD(W2,n2))) return NODE+1;
 if (!POS->wtm) GET_JOINT_DESCENDANT_LIST(G,W1,n1,W2,n2);
 // !POS->wtm implies this is only valid for verifiable trees
 else if (n1!=-1 && DATA_HAS_CHILD(W1,n1)) G[0]=W1[n1+1].move;
 else G[0]=W2[n2+1].move; mv=G[w++];
 SET_CHILD(NEW,NODE); NEW[++NODE].move=mv; LAST=NODE;
 MakeMoveCheck(POS,mv,FALSE); NODE=StepWinTree(POS,W1,W2,NEW,NODE);
 UnmakeMove(POS,mv,0);
 while ((mv=G[w++]))
 {if (NEW[LAST].data&(1U<<31)) NEW[LAST].data|=(1U<<30);
  else NEW[LAST].data|=NODE; NEW[NODE].move=mv; LAST=NODE;
  MakeMoveCheck(POS,mv,FALSE); NODE=StepWinTree(POS,W1,W2,NEW,NODE);
  UnmakeMove(POS,mv,0);} return NODE;}

void do_splice(typePOS *POS,WIN_NODE *W1,WIN_NODE *W2,boolean SAVE)
{WIN_NODE *SPL; uint64 u,sz,asz; char STR[256];
 init_position(POS,"startpos"); sz=(DATA_GET_NODE(W1,0)+DATA_GET_NODE(W2,0));
 sz+=sz/8+0x10000; HASH_LIMIT=2U<<BSR(sz);
 asz=((uint64) HASH_LIMIT)*sizeof(tTRANS3); HASH_TABLE=malloc(asz);
 printf("HASH_LIMIT is %d [%lldmb]\n",HASH_LIMIT,asz>>20);
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].key=HASH_TABLE[u].main=0;
 for (u=0;u<HASH_LIMIT;u++) HASH_TABLE[u].t1=HASH_TABLE[u].t2=-1;
 SPL=malloc(sz*sizeof(WIN_NODE)); asz=((uint64) sz)*sizeof(WIN_NODE);
 printf("Allocating for %lld records [%lldmb]\n",sz,asz>>20);
 for (u=0;u<sz;u++) SPL[u].data=SPL[u].move=0;
 fill_hash_table(POS,W1,0,0); printf("Filled tree 1\n");
 fill_hash_table(POS,W2,0,1); printf("Filled tree 2\n");
 SPL[0].data=(1U<<30)|StepWinTree(POS,W1,W2,SPL,0);
 strcpy(STR,"PROOFS/SPLICED.proof"); sz=DATA_GET_NODE(SPL,0);
 printf("New size is %lld (%lldmb), save to %s\n",sz,(sz*6)>>20,STR);
 if (SAVE) SaveWinTree(SPL,NULL,0,STR);}
