
#include "losing.h"

#define LOSS_TWIST 0x913bb468de01037fULL
#define WIN_TWIST 0xc066034f796185ebULL

static HASH *BOOK_HASH; static uint32 BOOK_HASH_MASK;

static uint32 NewBookHash(uint64 Z,uint32 NODE)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return BOOK_HASH[k].node; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
  BOOK_HASH[k].key=Z; BOOK_HASH[k].node=NODE; return 0;}

static uint32 IsKnownResult(uint64 Z)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return BOOK_HASH[k].node; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
 return FALSE;}

static void InstallKnownResult(uint64 Z,uint32 n)
{uint32 k=Z&BOOK_HASH_MASK; uint64 u; if (!k) k++;
 while ((u=BOOK_HASH[k].key))
 {if (u==Z) return; k++; if (k==(BOOK_HASH_MASK+1)) k=1;}
  BOOK_HASH[k].key=Z; BOOK_HASH[k].node=n;}

static HASH *ALL_HASH; static uint32 ALL_HASH_MASK;

static void MakeAllHash(uint32 n,uint64 Z)
{uint32 k=Z&ALL_HASH_MASK;
 while (ALL_HASH[k].key) {k++; if (k==(ALL_HASH_MASK+1)) k=0;}
 ALL_HASH[k].key=Z; ALL_HASH[k].node=n;}

////////////////////////////////////////////////////////////////////////

typedef struct {uint32 n,good,bad;} SORTER;
static int sort_compare(const void *x, const void *y) // increasing good
{SORTER A=*((SORTER*) x); SORTER B=*((SORTER*) y);
 if (A.good<B.good) return -1; if (A.good>B.good) return 1;
 if (A.bad>B.bad) return -1; if (A.bad<B.bad) return 1; return 0;}

static void SortChildren(PN_NODE *T,uint32 n)
{SORTER S[256]; int i,u=0; uint32 c=T[n].child;
 S[u].bad=T[c].bad; S[u].good=T[c].good; S[u++].n=c;
 while (T[c].sibling)
 {c=T[c].sibling; S[u].bad=T[c].bad; S[u].good=T[c].good; S[u++].n=c;}
 qsort(S,u,sizeof(SORTER),sort_compare); c=S[0].n; T[n].child=S[0].n;
 for (i=1;i<u;i++) {T[c].sibling=S[i].n; c=S[i].n;} T[c].sibling=0;}

////////////////////////////////////////////////////////////////////////

static void BacktrackLoop(PN_NODE *T,uint32 node);

static void win_installer(PN_NODE *T,uint32 n)
{uint32 u,k; uint64 Z,z; Z=T[n].hash; InstallKnownResult(Z^WIN_TWIST,n);
 T[n].loop=0; T[n].trans=n; T[n].bad=0; T[n].good=MY_INFoo; k=Z&ALL_HASH_MASK;
 while ((z=ALL_HASH[k].key))
 {u=ALL_HASH[k].node; k++; if (k==(ALL_HASH_MASK+1)) k=0;
  if ((z==Z) && u!=n)
  {if (!T[u].good) continue; T[u].bad=0; T[u].good=MY_INFoo;
   T[u].flags|=4; T[u].trans=n; T[u].loop=0; T[u].child=0;
   BacktrackLoop(T,T[u].parent);}}}

static void loss_installer(PN_NODE *T,uint32 n)
{uint32 u,k; uint64 Z,z; Z=T[n].hash; InstallKnownResult(Z^LOSS_TWIST,n);
 T[n].loop=0; T[n].trans=n; T[n].good=0; T[n].bad=MY_INFoo; k=Z&ALL_HASH_MASK;
 while ((z=ALL_HASH[k].key))
 {u=ALL_HASH[k].node; k++; if (k==(ALL_HASH_MASK+1)) k=0;
  if ((z==Z) && u!=n)
  {if (!T[u].bad) continue; T[u].good=0; T[u].bad=MY_INFoo;
   T[u].flags|=4; T[u].trans=n; T[u].loop=0; T[u].child=0;
   BacktrackLoop(T,T[u].parent);}}}

static boolean UpdateNode(PN_NODE *T,uint32 n)
{uint32 c,v,w=T[n].good,l=T[n].bad; if (!T[n].child) return TRUE;
 SortChildren(T,n); if (!w || !l) return FALSE;
 c=T[n].child; v=T[c].bad; T[n].bad=T[c].good;
 while (T[c].sibling) {c=T[c].sibling; v+=T[c].bad;if (v>MY_INFoo) v=MY_INFoo;}
 if (w && !v && !(T[n].flags&16) &&
     !IsKnownResult(T[n].hash^LOSS_TWIST)) loss_installer(T,n);
 if (l && !T[n].bad && (T[n].flags&16) &&
     !IsKnownResult(T[n].hash^WIN_TWIST)) win_installer(T,n);
 T[n].good=v; if (v==w && l==T[n].bad) return FALSE; return TRUE;}

static void BacktrackLoop(PN_NODE *T,uint32 node)
{uint32 n,k; if (!node) return; n=T[node].loop?T[node].trans:node;
 if (UpdateNode(T,n)) BacktrackLoop(T,T[n].parent);
 k=T[n].loop; if (!k) return;
 while (k!=n && k)
 {uint32 g=T[k].good,b=T[k].bad; T[k].good=T[n].good; T[k].bad=T[n].bad;
  if (g!=T[n].good || b!=T[n].bad) BacktrackLoop(T,T[k].parent); k=T[k].loop;}}

static uint32 iTreeMove(PN_NODE *T,uint16 mv,uint32 n)
{uint32 x=T[n].child; if (!n) return 0;
 do {if (T[x].move==mv) return x;} while ((x=T[x].sibling)); return 0;}

static uint32 TreeMove(PN_NODE *T,uint16 mv,uint32 n)
{uint32 r=iTreeMove(T,mv,n); if (r==0) return 0;
 if (T[r].loop || T[r].flags&4) return T[r].trans; return r;}

////////////////////////////////////////////////////////////////////////

#define SET_CHILD(A,x) (A[x].data|=(1U<<30))
#define DATA_IS_TRANS(A,x) (((A[x].data&(1U<<31)))==(1U<<31))
#define DATA_HAS_CHILD(A,x)\
  (((A[x].data&(3U<<30)))==(1U<<30) && ((A[x].data&0x3fffffff)!=0x3fffffff))
#define TRANS_AND_SIBLING(A,x) (((A[x].data&(3U<<30)))==(3U<<30))
#define DATA_GET_NODE(A,x) (A[x].data&0x3fffffff)
#define DATA_GET_NODEns(A,x) \
  (((A[x].data&0x3fffffff)!=0x3fffffff) ? A[x].data&0x3fffffff : \
   (((A[x].data&(1U<<30))==(1U<<30)) ? ((x)+1) : 0))
#define NEXT_SIBLING(A,x) ((TRANS_AND_SIBLING(A,x)) ? ((x)+1) : \
			   (DATA_IS_TRANS(A,x) ? 0 : DATA_GET_NODEns(A,x)))


static uint32 WinTreeMove(WIN_NODE *W,uint16 mv,uint32 n)
{uint32 c=n+1; if (!DATA_HAS_CHILD(W,n)) return -1;
 do {if (W[c].move==mv) return DATA_IS_TRANS(W,c)?DATA_GET_NODE(W,c):c;}
 while ((c=NEXT_SIBLING(W,c))); return -1;}

static uint32 cnt=0,ok=0;
static void NewNode(typePOS *POS,uint16 mv,PN_NODE *OLD,WIN_NODE *WT,
		    uint32 oldpos,PN_NODE *NEW,uint32 parent)
{uint32 r,t,c,node=NEW[0].size++; uint64 HASH=POS->DYN->HASH; uint16 ML[256];
 if (!(node&1023)) printf("NewNode %d\n",node); MakeAllHash(node,HASH);
 NEW[node].good=1; NEW[node].bad=1; NEW[node].hash=HASH;
 NEW[node].parent=parent; NEW[node].sibling=NEW[node].child=0;
 NEW[node].move=mv; NEW[node].size=NEW[node].killer=0; NEW[node].size=0;
 NEW[node].flags=POS->wtm?16:0; NEW[node].trans=node; NEW[node].loop=0;
 if (!NEW[parent].child) NEW[parent].child=node;
 else {c=NEW[parent].child; while (NEW[c].sibling) c=NEW[c].sibling;
       NEW[c].sibling=node;} // worry patt/rep, TBs/oppB are covered in search
 if (!GenMoves(POS,ML)) // stalemate, combines FICS/International
 {int wc=POPCNT(wBitboardOcc),bc=POPCNT(bBitboardOcc);
  if (POS->wtm && wc<bc) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}
  if (POS->wtm && wc>=bc) {NEW[node].good=0; NEW[node].bad=MY_INFoo; return;}
  if (!POS->wtm) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}}
 for (r=4;r<=POS->DYN->rev;r+=2) if (((POS->DYN-r)->HASH)==HASH) // rep
 {if (POS->wtm) {NEW[node].good=0; NEW[node].bad=MY_INFoo; return;}
  if (!POS->wtm) {NEW[node].bad=0; NEW[node].good=MY_INFoo; return;}}
 if (POS->wtm && (t=IsKnownResult(HASH^WIN_TWIST)))
 {NEW[node].good=MY_INFoo; NEW[node].bad=0;
  NEW[node].flags|=4; NEW[node].trans=t; NEW[node].loop=0; return;} 
 if (!POS->wtm && (t=IsKnownResult(HASH^LOSS_TWIST)))
 {NEW[node].bad=MY_INFoo; NEW[node].good=0;
  NEW[node].flags|=4; NEW[node].trans=t; NEW[node].loop=0; return;} 
 if (known_find(HASH)!=-1)
 {(POS->wtm)?win_installer(NEW,node):loss_installer(NEW,node); ok++; return;}
 if (POS->DYN->rev==0 && (t=NewBookHash(HASH,node)))
 {NEW[node].trans=t; NEW[node].good=NEW[t].good; NEW[node].bad=NEW[t].bad;
  NEW[node].loop=NEW[t].loop?NEW[t].loop:t; NEW[t].loop=node;}
 if (OLD && oldpos && !OLD[oldpos].child && // oldpos is terminal, unresolved
     ((POS->wtm && !OLD[oldpos].bad) || (!POS->wtm && !OLD[oldpos].good)))
 {if (POS->wtm) NEW[node].good=MY_INFoo; else NEW[node].bad=MY_INFoo;
  if (!OLD[oldpos].killer) cnt++; OLD[oldpos].killer=1; NEW[node].flags|=1;}
 if (WT && oldpos!=-1 && !DATA_HAS_CHILD(WT,oldpos))
 {if (POS->wtm) NEW[node].good=MY_INFoo; else NEW[node].bad=MY_INFoo;
  cnt++; NEW[node].flags|=1;}
 if (NEW[node].flags&1) printf("unKNOWN 0x%016llx\n",HASH);}

static void recurse
(typePOS *POS,PN_NODE *OLD,uint32 oldpos,PN_NODE *NEW,uint32 cur)
{uint32 n,k,c,h,x; uint16 ML[256]; n=GenMoves(POS,ML);
 for (k=0;k<n;k++)
 {MakeMoveCheck(POS,ML[k],FALSE); h=TreeMove(OLD,ML[k],oldpos);
  NewNode(POS,ML[k],OLD,NULL,h,NEW,cur); UnmakeMove(POS,ML[k],TRUE);}
 BacktrackLoop(NEW,cur); // this should "sort" children?
 if (!NEW[cur].good || !NEW[cur].bad) return; c=OLD[oldpos].child;
 do {if (!POS->wtm || !OLD[c].good)
     {MakeMoveCheck(POS,OLD[c].move,FALSE); h=TreeMove(NEW,OLD[c].move,cur);
      if (OLD[c].loop || OLD[c].flags&4) x=OLD[c].trans; else x=c;
      if (NEW[h].good && NEW[h].bad && OLD[x].child && !NEW[h].child)
	recurse(POS,OLD,x,NEW,h);
      UnmakeMove(POS,OLD[c].move,TRUE);}} while ((c=OLD[c].sibling));}

void recurse_wintree
(typePOS *POS,WIN_NODE *OLD,uint32 oldpos,PN_NODE *NEW,uint32 cur)
{uint32 n,k,c,h,x; uint16 ML[256]; n=GenMoves(POS,ML);
  for (k=0;k<n;k++)
  {MakeMoveCheck(POS,ML[k],FALSE); h=WinTreeMove(OLD,ML[k],oldpos);
   NewNode(POS,ML[k],NULL,OLD,h,NEW,cur); UnmakeMove(POS,ML[k],TRUE);}
  BacktrackLoop(NEW,cur); // this should "sort" children?
  if (!NEW[cur].good || !NEW[cur].bad) return;
  c=DATA_HAS_CHILD(OLD,oldpos)?oldpos+1:0;
  do {MakeMoveCheck(POS,OLD[c].move,FALSE); h=TreeMove(NEW,OLD[c].move,cur);
      x=DATA_IS_TRANS(OLD,c)?DATA_GET_NODE(OLD,c):c;
      if (NEW[h].good && NEW[h].bad && DATA_HAS_CHILD(OLD,x) && !NEW[h].child)
	recurse_wintree(POS,OLD,x,NEW,h); // probably avoid 0x3fffffff data?
      UnmakeMove(POS,OLD[c].move,TRUE);} while ((c=NEXT_SIBLING(OLD,c)));}

static void SaveNode(FILE *F,PN_NODE *P)
{fwrite(&(P->good),4,1,F); fwrite(&(P->bad),4,1,F); fwrite(&(P->size),4,1,F);
 fwrite(&(P->hash),8,1,F); fwrite(&(P->trans),4,1,F); fwrite(&(P->loop),4,1,F);
 fwrite(&(P->parent),4,1,F); fwrite(&(P->sibling),4,1,F);
 fwrite(&(P->child),4,1,F); fwrite(&(P->move),2,1,F);
 fwrite(&(P->killer),1,1,F); fwrite(&(P->flags),1,1,F);}

static void SaveNewBook(PN_NODE *T,char *fn)
{uint32 u; FILE *F; F=fopen(fn,"w"); if (!F) {printf("Save failed\n"); return;}
 printf("Saving %s with %d positions\n",fn,T[0].size); fflush(stdout);
 for (u=0;u<T[0].size;u++) SaveNode(F,T+u); fclose(F);}

void tree_build(typePOS *POS,PN_NODE *OLD,char *ml)
{PN_NODE *NEW; uint64 all,u,sz=OLD[0].size;
 NEW=CALLOC(sz+0x100000,sizeof(PN_NODE));
 NEW[0].child=NEW[0].parent=NEW[0].sibling=0; NEW[0].good=NEW[0].bad=0;
 NEW[0].loop=NEW[0].trans=0; NEW[0].hash=0;
 NEW[0].move=NEW[0].flags=NEW[0].killer=0; init_position(POS,"startpos");
 NEW[0].size=2; NEW[1].good=NEW[1].bad=1; NEW[1].hash=POS->DYN->HASH;
 NEW[1].parent=NEW[1].child=NEW[1].sibling=0; NEW[1].trans=1; NEW[1].loop=0;
 NEW[1].size=0; NEW[1].move=0; NEW[1].flags=16; NEW[1].killer=0;
 all=2<<BSR(8192+sz); if (all<(1<<20)) all=(1<<20); BOOK_HASH_MASK=all-1;
 BOOK_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 all=2<<BSR(8192+sz+sz/2); if (all<(1<<20)) all=(1<<20); ALL_HASH_MASK=all-1;
 ALL_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) ALL_HASH[u].key=0;
 parse_moves(POS,ml); typeDYNAMIC *D; D=POS->DYN_ROOT+1; uint32 w=2,n=1;
 while (D!=POS->DYN && n)
 {D++; if (OLD[n].loop) n=OLD[n].trans; n=TreeMove(OLD,D->mv,n);}
 if (!n) {printf("Failure to align moves\n"); return;} D=POS->DYN_ROOT+1;
 while (D!=POS->DYN)
 {D++; NEW[w].good=1; NEW[w].bad=1; NEW[w].move=D->mv; NEW[w].parent=w-1;
  NEW[w-1].child=w; NEW[w].sibling=0; NEW[w].trans=w; NEW[w].loop=0;
  NEW[w].child=0; NEW[w].killer=0; NEW[w].hash=D->HASH; NEW[w].size=0;
  NEW[w].flags=(w&1)?16:0; w++;}
 NEW[0].size=w; recurse(POS,OLD,n,NEW,w-1); free(ALL_HASH); free(BOOK_HASH);
 printf("Found %d terminal pos, %d lack KNOWN\n",ok+cnt,cnt); fflush(stdout);
 SaveNewBook(NEW,"BOOKS/SIMPLIFIED.book"); free(NEW);}

void tree_build_wintree(typePOS *POS,WIN_STRUCT *OLD,char *ml)
{PN_NODE *NEW; uint64 all,u,sz=OLD->W[0].data&0x3fffffff;
 NEW=CALLOC(sz+0x100000,sizeof(PN_NODE));
 NEW[0].child=NEW[0].parent=NEW[0].sibling=0; NEW[0].good=NEW[0].bad=0;
 NEW[0].loop=NEW[0].trans=0; NEW[0].hash=0;
 NEW[0].move=NEW[0].flags=NEW[0].killer=0; init_position(POS,"startpos");
 NEW[0].size=2; NEW[1].good=NEW[1].bad=1; NEW[1].hash=POS->DYN->HASH;
 NEW[1].parent=NEW[1].child=NEW[1].sibling=0; NEW[1].trans=1; NEW[1].loop=0;
 NEW[1].size=0; NEW[1].move=0; NEW[1].flags=16; NEW[1].killer=0;
 all=2<<BSR(8192+sz); if (all<(1<<20)) all=(1<<20); BOOK_HASH_MASK=all-1;
 BOOK_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) BOOK_HASH[u].key=0;
 all=2<<BSR(8192+sz+sz/2); if (all<(1<<20)) all=(1<<20); ALL_HASH_MASK=all-1;
 ALL_HASH=CALLOC(all,sizeof(HASH)); for (u=0;u<all;u++) ALL_HASH[u].key=0;
 parse_moves(POS,ml); typeDYNAMIC *D; D=POS->DYN_ROOT+1; uint32 w=2,n=0;
 for (u=0;u<OLD->W[0].move;u++) // needs to handle prolog correctly
 {D++; if (D->mv!=OLD->MOVE_LIST[u]) {printf("Prolog failure\n"); return;}}
 while (D!=POS->DYN && n!=-1) {D++; n=WinTreeMove(OLD->W,D->mv,n);}
 if (n==-1) {printf("Failure to align moves\n"); return;} D=POS->DYN_ROOT+1;
 while (D!=POS->DYN)
 {D++; NEW[w].good=1; NEW[w].bad=1; NEW[w].move=D->mv; NEW[w].parent=w-1;
  NEW[w-1].child=w; NEW[w].sibling=0; NEW[w].trans=w; NEW[w].loop=0;
  NEW[w].child=0; NEW[w].killer=0; NEW[w].hash=D->HASH; NEW[w].size=0;
  NEW[w].flags=(w&1)?16:0; w++;}
 NEW[0].size=w; recurse_wintree(POS,OLD->W,n,NEW,w-1);
 free(ALL_HASH); free(BOOK_HASH);
 printf("Found %d terminal pos, %d lack KNOWN\n",ok+cnt,cnt); fflush(stdout);
 SaveNewBook(NEW,"BOOKS/SIMPLIFIED.book"); free(NEW);}
